data = (
'poss',    # 0x00
'pong',    # 0x01
'poj',    # 0x02
'poc',    # 0x03
'pok',    # 0x04
'pot',    # 0x05
'pop',    # 0x06
'poh',    # 0x07
'pwa',    # 0x08
'pwag',    # 0x09
'pwagg',    # 0x0a
'pwags',    # 0x0b
'pwan',    # 0x0c
'pwanj',    # 0x0d
'pwanh',    # 0x0e
'pwad',    # 0x0f
'pwal',    # 0x10
'pwalg',    # 0x11
'pwalm',    # 0x12
'pwalb',    # 0x13
'pwals',    # 0x14
'pwalt',    # 0x15
'pwalp',    # 0x16
'pwalh',    # 0x17
'pwam',    # 0x18
'pwab',    # 0x19
'pwabs',    # 0x1a
'pwas',    # 0x1b
'pwass',    # 0x1c
'pwang',    # 0x1d
'pwaj',    # 0x1e
'pwac',    # 0x1f
'pwak',    # 0x20
'pwat',    # 0x21
'pwap',    # 0x22
'pwah',    # 0x23
'pwae',    # 0x24
'pwaeg',    # 0x25
'pwaegg',    # 0x26
'pwaegs',    # 0x27
'pwaen',    # 0x28
'pwaenj',    # 0x29
'pwaenh',    # 0x2a
'pwaed',    # 0x2b
'pwael',    # 0x2c
'pwaelg',    # 0x2d
'pwaelm',    # 0x2e
'pwaelb',    # 0x2f
'pwaels',    # 0x30
'pwaelt',    # 0x31
'pwaelp',    # 0x32
'pwaelh',    # 0x33
'pwaem',    # 0x34
'pwaeb',    # 0x35
'pwaebs',    # 0x36
'pwaes',    # 0x37
'pwaess',    # 0x38
'pwaeng',    # 0x39
'pwaej',    # 0x3a
'pwaec',    # 0x3b
'pwaek',    # 0x3c
'pwaet',    # 0x3d
'pwaep',    # 0x3e
'pwaeh',    # 0x3f
'poe',    # 0x40
'poeg',    # 0x41
'poegg',    # 0x42
'poegs',    # 0x43
'poen',    # 0x44
'poenj',    # 0x45
'poenh',    # 0x46
'poed',    # 0x47
'poel',    # 0x48
'poelg',    # 0x49
'poelm',    # 0x4a
'poelb',    # 0x4b
'poels',    # 0x4c
'poelt',    # 0x4d
'poelp',    # 0x4e
'poelh',    # 0x4f
'poem',    # 0x50
'poeb',    # 0x51
'poebs',    # 0x52
'poes',    # 0x53
'poess',    # 0x54
'poeng',    # 0x55
'poej',    # 0x56
'poec',    # 0x57
'poek',    # 0x58
'poet',    # 0x59
'poep',    # 0x5a
'poeh',    # 0x5b
'pyo',    # 0x5c
'pyog',    # 0x5d
'pyogg',    # 0x5e
'pyogs',    # 0x5f
'pyon',    # 0x60
'pyonj',    # 0x61
'pyonh',    # 0x62
'pyod',    # 0x63
'pyol',    # 0x64
'pyolg',    # 0x65
'pyolm',    # 0x66
'pyolb',    # 0x67
'pyols',    # 0x68
'pyolt',    # 0x69
'pyolp',    # 0x6a
'pyolh',    # 0x6b
'pyom',    # 0x6c
'pyob',    # 0x6d
'pyobs',    # 0x6e
'pyos',    # 0x6f
'pyoss',    # 0x70
'pyong',    # 0x71
'pyoj',    # 0x72
'pyoc',    # 0x73
'pyok',    # 0x74
'pyot',    # 0x75
'pyop',    # 0x76
'pyoh',    # 0x77
'pu',    # 0x78
'pug',    # 0x79
'pugg',    # 0x7a
'pugs',    # 0x7b
'pun',    # 0x7c
'punj',    # 0x7d
'punh',    # 0x7e
'pud',    # 0x7f
'pul',    # 0x80
'pulg',    # 0x81
'pulm',    # 0x82
'pulb',    # 0x83
'puls',    # 0x84
'pult',    # 0x85
'pulp',    # 0x86
'pulh',    # 0x87
'pum',    # 0x88
'pub',    # 0x89
'pubs',    # 0x8a
'pus',    # 0x8b
'puss',    # 0x8c
'pung',    # 0x8d
'puj',    # 0x8e
'puc',    # 0x8f
'puk',    # 0x90
'put',    # 0x91
'pup',    # 0x92
'puh',    # 0x93
'pweo',    # 0x94
'pweog',    # 0x95
'pweogg',    # 0x96
'pweogs',    # 0x97
'pweon',    # 0x98
'pweonj',    # 0x99
'pweonh',    # 0x9a
'pweod',    # 0x9b
'pweol',    # 0x9c
'pweolg',    # 0x9d
'pweolm',    # 0x9e
'pweolb',    # 0x9f
'pweols',    # 0xa0
'pweolt',    # 0xa1
'pweolp',    # 0xa2
'pweolh',    # 0xa3
'pweom',    # 0xa4
'pweob',    # 0xa5
'pweobs',    # 0xa6
'pweos',    # 0xa7
'pweoss',    # 0xa8
'pweong',    # 0xa9
'pweoj',    # 0xaa
'pweoc',    # 0xab
'pweok',    # 0xac
'pweot',    # 0xad
'pweop',    # 0xae
'pweoh',    # 0xaf
'pwe',    # 0xb0
'pweg',    # 0xb1
'pwegg',    # 0xb2
'pwegs',    # 0xb3
'pwen',    # 0xb4
'pwenj',    # 0xb5
'pwenh',    # 0xb6
'pwed',    # 0xb7
'pwel',    # 0xb8
'pwelg',    # 0xb9
'pwelm',    # 0xba
'pwelb',    # 0xbb
'pwels',    # 0xbc
'pwelt',    # 0xbd
'pwelp',    # 0xbe
'pwelh',    # 0xbf
'pwem',    # 0xc0
'pweb',    # 0xc1
'pwebs',    # 0xc2
'pwes',    # 0xc3
'pwess',    # 0xc4
'pweng',    # 0xc5
'pwej',    # 0xc6
'pwec',    # 0xc7
'pwek',    # 0xc8
'pwet',    # 0xc9
'pwep',    # 0xca
'pweh',    # 0xcb
'pwi',    # 0xcc
'pwig',    # 0xcd
'pwigg',    # 0xce
'pwigs',    # 0xcf
'pwin',    # 0xd0
'pwinj',    # 0xd1
'pwinh',    # 0xd2
'pwid',    # 0xd3
'pwil',    # 0xd4
'pwilg',    # 0xd5
'pwilm',    # 0xd6
'pwilb',    # 0xd7
'pwils',    # 0xd8
'pwilt',    # 0xd9
'pwilp',    # 0xda
'pwilh',    # 0xdb
'pwim',    # 0xdc
'pwib',    # 0xdd
'pwibs',    # 0xde
'pwis',    # 0xdf
'pwiss',    # 0xe0
'pwing',    # 0xe1
'pwij',    # 0xe2
'pwic',    # 0xe3
'pwik',    # 0xe4
'pwit',    # 0xe5
'pwip',    # 0xe6
'pwih',    # 0xe7
'pyu',    # 0xe8
'pyug',    # 0xe9
'pyugg',    # 0xea
'pyugs',    # 0xeb
'pyun',    # 0xec
'pyunj',    # 0xed
'pyunh',    # 0xee
'pyud',    # 0xef
'pyul',    # 0xf0
'pyulg',    # 0xf1
'pyulm',    # 0xf2
'pyulb',    # 0xf3
'pyuls',    # 0xf4
'pyult',    # 0xf5
'pyulp',    # 0xf6
'pyulh',    # 0xf7
'pyum',    # 0xf8
'pyub',    # 0xf9
'pyubs',    # 0xfa
'pyus',    # 0xfb
'pyuss',    # 0xfc
'pyung',    # 0xfd
'pyuj',    # 0xfe
'pyuc',    # 0xff
)



from typing import List

from ._PickupAreaSector import PickupAreaSector


class PickupArea(object):
    def __init__(
        self,           # type: PickupArea
        pickup_name,    # type: str
        sectors,        # type: List[PickupAreaSector]
    ):
        # type: (...) -> None

        self.pickup_name = pickup_name
        self.sectors = sectors



from typing import List

from mw_helper import ensure_iterable

from production.model import ProductionOrder
from ._OrderCondition import OrderCondition


class PodTypeCondition(OrderCondition):
    def __init__(self, pod_types):
        # type: (List[str]) -> None

        self.valid_pod_types = ensure_iterable(pod_types)

    def evaluate(self, order):
        # type: (ProductionOrder) -> bool

        return order.pod_type in self.valid_pod_types



import os
import iso8601
import numpy as np

from typing import List
from os import listdir
from os.path import isfile, join

# config:
log_directory = "C:\\Users\\estil\\Downloads\\203\\data\\server\\bundles\\tablemgr"
start_date = iso8601.parse_date("2023-06-12 09:00:00")
end_date = iso8601.parse_date("2023-06-13 03:00:00")


class LogLine(object):
    def __init__(self, thread_number, log_date, message, line_content):
        self.thread_number = thread_number
        self.log_date = log_date
        self.message = message
        self.line_content = line_content


def read_log_files(log_files):
    file_data = ""
    for file_name in log_files:
        with open(file_name, "rb") as f:
            file_data += f.read()

    return file_data.decode("utf-8")


def get_thread_number(line):
    return line.split("-")[4].strip()


def get_order_id(line_content):
    #  type: (str) -> int
    index = line_content.index("\n")
    if index < 0:
        return -1
    index2 = line_content.index("\n", index + 1)
    order_header_line = line_content[index + 1:index2]
    comps = order_header_line.split(",")
    data = comps[0].split(":")
    try:
        return int(data[1].strip())
    except:
        return -1


def get_message(line_content):
    try:
        index = line_content.index("Thread-")
    except ValueError:
        index = line_content.index("MainThread")
    index2 = line_content.index(" - ", index)
    return line_content[index2 + 3:]


def get_log_date(line_content):
    return iso8601.parse_date(line_content[0:23])


def get_log_lines(log_data):
    lines = []
    line_content = ""
    first_line = None
    for line in log_data.split('\n'):
        line_start = line[0:23]
        try:
            line_date = iso8601.parse_date(line_start)
            if line_date < start_date or line_date > end_date:
                continue

            if line_content != "":
                lines.append(LogLine(
                    get_thread_number(first_line),
                    get_log_date(first_line),
                    get_message(line_content),
                    line_content
                ))
                line_content = ""

            first_line = line
        except:
            pass

        line_content += line + "\n"

    if line_content != "":
        lines.append(LogLine(
            get_thread_number(first_line),
            get_log_date(first_line),
            get_message(line_content),
            line_content
        ))

    return lines


def separate_lines_by_thread(log_lines):
    # type: (List[LogLine]) -> List[List[LogLine]]
    line_dict = {}
    for line in log_lines:
        if line.thread_number not in line_dict:
            line_dict[line.thread_number] = []
        line_dict[line.thread_number].append(line)

    return line_dict.values()


def separate_lines_by_order(log_lines):
    # type: (List[LogLine]) -> List[List[LogLine]]
    line_dict = {}
    for line in log_lines:
        if line.order_id not in line_dict:
            line_dict[line.order_id] = []
        line_dict[line.order_id].append(line)

    return line_dict


def get_connection_id(message):
    index = message.index("c_void_p(")
    index2 = message.index(")", index)
    return message[index + 9:index2]


def main():
    log_files = [os.path.join(log_directory, f) for f in listdir(log_directory) if isfile(join(log_directory, f)) and ".log" in f]
    log_lines = get_log_lines(read_log_files(log_files))

    separated_lines = separate_lines_by_thread(log_lines)

    opened_connections = {}
    connections_with_transaction = {}
    message_times = {}
    message_max_times = {}

    query_times = {}
    query_max_times = {}

    start_time = None
    message = None
    start_line = None

    query_name = None
    start_list_table = None

    print("Start Logging Analisys...")
    for thread_logs in separated_lines:
        for line in thread_logs:
            line.message = line.message.strip()

            if line.message.startswith("Opening connection: "):
                opened_connections[get_connection_id(line.message)] = ""

            elif line.message.startswith("Closing connection: "):
                connection_id = get_connection_id(line.message)
                if connection_id not in opened_connections:
                    print("Connection {} not opened".format(connection_id))
                else:
                    del opened_connections[connection_id]

            elif line.message.startswith("Starting a transaction: "):
                connections_with_transaction[get_connection_id(line.message)] = ""

            elif line.message.startswith("Ending a transaction: "):
                connection_id = get_connection_id(line.message)
                if connection_id not in connections_with_transaction:
                    print("Transaction for connection {} not opened".format(connection_id))
                else:
                    del connections_with_transaction[connection_id]

            elif line.message.startswith("Handling: "):
                start_time = line.log_date
                start_line = line
                message = line.message[10:].strip().split('[')[0]

            elif line.message.startswith("Successfully processed: "):
                end_time = line.log_date

                if end_time is not None and start_time is not None:
                    elapsed_time = (end_time - start_time).total_seconds()
                    if elapsed_time > 60:
                        print(start_line)

                    if elapsed_time > 0:
                        if message not in message_times:
                            message_times[message] = []

                        if message not in message_max_times:
                            message_max_times[message] = elapsed_time
                        elif elapsed_time > message_max_times[message]:
                            message_max_times[message] = elapsed_time
                        message_times[message].append(elapsed_time)
                start_time = None

            elif line.message.startswith("Executing"):
                start_list_table = line.log_date
                query_name = line.message.split('-')[-1]

            elif line.message.startswith("Executed"):
                end_time_list_table = line.log_date
                if end_time_list_table is not None and start_list_table is not None:
                    elapsed_time = (end_time_list_table - start_list_table).total_seconds()
                    if elapsed_time > 0:
                        if query_name not in query_times:
                            query_times[query_name] = []

                        if query_name not in query_max_times:
                            query_max_times[query_name] = elapsed_time
                        elif elapsed_time > query_max_times[query_name]:
                            query_max_times[query_name] = elapsed_time
                        query_times[query_name].append(elapsed_time)
                start_list_table = None

        if len(opened_connections) > 0:
            print("Unclosed connections: {}".format(opened_connections))

        if len(connections_with_transaction) > 0:
            print("Unfinished transactions: {}".format(opened_connections))

    messages = message_times.keys()
    messages.sort()
    for message in messages:
        times = message_times[message]
        a = np.array(times)
        p = np.percentile(a, 95)  # return 95th percentile
        print("{:25} - [{:>4}]: {:>6.3f} / {:>6.3f}".format(message, len(times), np.average(times), p))
    print("")

    messages = query_times.keys()
    messages.sort()
    for query in messages:
        times = query_times[query]
        a = np.array(times)
        p = np.percentile(a, 98)  # return 95th percentile
        if float(p) > 0.020:
            print("{:35} - [{:>4}]: {:>6.3f} / {:>6.3f}".format(query, len(times), np.average(times), p))
    print("")

    for message in message_max_times:
        print("{:40}: {:>6.3f}".format(message, message_max_times[message]))
    print("")

    for query in query_max_times:
        if float(query_max_times[query]) > 0.020:
            print("{:40}: {:>6.3f}".format(query, query_max_times[query]))
    print("")


main()



from production.box import JitAddTagBox
from production.model import TagNotFoundException
from production.treebuilder import BoxConfiguration

from ._BoxBuilder import BoxBuilder


class JitAddTagBoxBuilder(BoxBuilder):
    def __init__(self, product_repository, loggers):
        super(JitAddTagBoxBuilder, self).__init__(loggers)
        self.product_repository = product_repository

    def build(self, config):
        # type: (BoxConfiguration) -> JitAddTagBox
        if u"Tag" not in config.extra_config:
            raise TagNotFoundException("JitAddTagBoxBuilder without tag: {}".format(config.name))

        tag = config.extra_config[u"Tag"]

        return JitAddTagBox(config.name,
                            config.sons,
                            self.product_repository,
                            tag,
                            self.get_extra(config, "JitLines", []),
                            self.get_logger(config))



# -*- coding: utf-8 -*-
from enum import IntEnum


class TransferType(IntEnum):
    INITIAL_AMOUNT = 1
    TRANSFER_SKIM = 2
    TRANSFER_CASH_IN = 3
    TRANSFER_CASH_OUT = 4
    FINAL_AMOUNT = 5
    DECLARED_AMOUNT = 6



from typing import Optional
from xml.etree import cElementTree as eTree

from sysactions import (
    UiOutOfSync,
    get_model,
    get_posot,
    get_current_order,
)

from actions import logger


def must_sync_ui(
    pos_id,             # type: str
    different_than,     # type: str
):
    # type: (...) -> None

    order = _get_order_picture(pos_id=pos_id)

    different_than = different_than.upper()
    state = order.get("state")
    if state != different_than:
        logger.info("Required Order Status: {}. Model Order Status: {}".format(
            different_than,
            state,
        ))

        raise UiOutOfSync(
            pos_id=pos_id,
            custom_message="$ERROR_ORDER_NOT_SYNCHRONIZED",
        )


def sync_ui_and_model(
    pos_id,     # type: str
):
    # type: (...) -> None

    raise UiOutOfSync(
        pos_id=pos_id,
        reset_model=True,
        custom_message="$ERROR_ORDER_NOT_SYNCHRONIZED",
    )


def _get_order_picture(
    pos_id,     # type: Optional[str]
):
    # type: (...) -> Optional[eTree.Element]

    model = get_model(posid=pos_id)
    pos_ot = get_posot(model=model)

    current_order = get_current_order(model=model)
    order_id = current_order.get("orderId")
    order_xml = eTree.XML(
        pos_ot.orderPicture(
            posid=pos_id,
            orderid=order_id,
        )
    )
    return order_xml.find("Order") if order_xml.find("Order") else order_xml



# acn.py - functions for handling Australian Company Numbers (ACNs)
#
# Copyright (C) 2016 Vincent Bastos
# Copyright (C) 2016 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""ACN (Australian Company Number).

The Australian Company Number (ACN) is a company identifier issued by the
Australian Securities and Investments Commission.

More information:

* https://en.wikipedia.org/wiki/Australian_Company_Number

>>> validate('004 085 616')
'004085616'
>>> validate('010 499 966')
'010499966'
>>> validate('999 999 999')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> format('004085616')
'004 085 616'
>>> to_abn('002 724 334')
'43002724334'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' ').strip()


def calc_check_digit(number):
    """Calculate the checksum."""
    return str((sum(int(n) * (i - 8) for i, n in enumerate(number))) % 10)


def validate(number):
    """Check if the number is a valid ACN. This checks the length, formatting
    and check digit."""
    number = compact(number)
    if not isdigits(number):
        raise InvalidFormat()
    if len(number) != 9:
        raise InvalidLength()
    if calc_check_digit(number) != number[-1]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid ACN."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return ' '.join((number[0:3], number[3:6], number[6:]))


def to_abn(number):
    """Convert the number to an Australian Business Number (ABN)."""
    from stdnum.au import abn
    number = compact(number)
    return abn.calc_check_digits(number) + number



from sysactions import action
from actions.order.close_active_order_and_pause_user import close_active_order_and_pause_user
from .. import logger


@action
def handle_power_off_screen_mobile(
    pos_id,     # type: str
):
    # type: (...) -> None

    logger.debug("Call handle power off screen mobile")
    close_active_order_and_pause_user(
        pos_id=pos_id,
        void_reason='POS_MOBILE_ON_PAUSE',
    )



# Copyright 2009-2017 Wander Lairson Costa
# Copyright 2009-2020 PyUSB contributors
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from ctypes import *
import usb.util
import sys
import logging
from usb._debug import methodtrace
import usb._interop as _interop
import usb._objfinalizer as _objfinalizer
import errno
import math
from usb.core import USBError, USBTimeoutError
import usb.libloader

__author__ = 'Wander Lairson Costa'

__all__ = [
            'get_backend',
            'LIBUSB_SUCCESS',
            'LIBUSB_ERROR_IO',
            'LIBUSB_ERROR_INVALID_PARAM',
            'LIBUSB_ERROR_ACCESS',
            'LIBUSB_ERROR_NO_DEVICE',
            'LIBUSB_ERROR_NOT_FOUND',
            'LIBUSB_ERROR_BUSY',
            'LIBUSB_ERROR_TIMEOUT',
            'LIBUSB_ERROR_OVERFLOW',
            'LIBUSB_ERROR_PIPE',
            'LIBUSB_ERROR_INTERRUPTED',
            'LIBUSB_ERROR_NO_MEM',
            'LIBUSB_ERROR_NOT_SUPPORTED',
            'LIBUSB_ERROR_OTHER',
            'LIBUSB_TRANSFER_COMPLETED',
            'LIBUSB_TRANSFER_ERROR',
            'LIBUSB_TRANSFER_TIMED_OUT',
            'LIBUSB_TRANSFER_CANCELLED',
            'LIBUSB_TRANSFER_STALL',
            'LIBUSB_TRANSFER_NO_DEVICE',
            'LIBUSB_TRANSFER_OVERFLOW',
        ]

_logger = logging.getLogger('usb.backend.libusb1')

# libusb.h

# transfer_type codes
# Control endpoint
_LIBUSB_TRANSFER_TYPE_CONTROL = 0,
# Isochronous endpoint
_LIBUSB_TRANSFER_TYPE_ISOCHRONOUS = 1
# Bulk endpoint
_LIBUSB_TRANSFER_TYPE_BULK = 2
# Interrupt endpoint
_LIBUSB_TRANSFER_TYPE_INTERRUPT = 3

# return codes

LIBUSB_SUCCESS = 0
LIBUSB_ERROR_IO = -1
LIBUSB_ERROR_INVALID_PARAM = -2
LIBUSB_ERROR_ACCESS = -3
LIBUSB_ERROR_NO_DEVICE = -4
LIBUSB_ERROR_NOT_FOUND = -5
LIBUSB_ERROR_BUSY = -6
LIBUSB_ERROR_TIMEOUT = -7
LIBUSB_ERROR_OVERFLOW = -8
LIBUSB_ERROR_PIPE = -9
LIBUSB_ERROR_INTERRUPTED = -10
LIBUSB_ERROR_NO_MEM = -11
LIBUSB_ERROR_NOT_SUPPORTED = -12
LIBUSB_ERROR_OTHER = -99

# map return codes to strings
_str_error_map = {
    LIBUSB_SUCCESS:'Success (no error)',
    LIBUSB_ERROR_IO:'Input/output error',
    LIBUSB_ERROR_INVALID_PARAM:'Invalid parameter',
    LIBUSB_ERROR_ACCESS:'Access denied (insufficient permissions)',
    LIBUSB_ERROR_NO_DEVICE:'No such device (it may have been disconnected)',
    LIBUSB_ERROR_NOT_FOUND:'Entity not found',
    LIBUSB_ERROR_BUSY:'Resource busy',
    LIBUSB_ERROR_TIMEOUT:'Operation timed out',
    LIBUSB_ERROR_OVERFLOW:'Overflow',
    LIBUSB_ERROR_PIPE:'Pipe error',
    LIBUSB_ERROR_INTERRUPTED:'System call interrupted (perhaps due to signal)',
    LIBUSB_ERROR_NO_MEM:'Insufficient memory',
    LIBUSB_ERROR_NOT_SUPPORTED:'Operation not supported or unimplemented on this platform',
    LIBUSB_ERROR_OTHER:'Unknown error'
}

# map return code to errno values
_libusb_errno = {
    0:None,
    LIBUSB_ERROR_IO:errno.__dict__.get('EIO', None),
    LIBUSB_ERROR_INVALID_PARAM:errno.__dict__.get('EINVAL', None),
    LIBUSB_ERROR_ACCESS:errno.__dict__.get('EACCES', None),
    LIBUSB_ERROR_NO_DEVICE:errno.__dict__.get('ENODEV', None),
    LIBUSB_ERROR_NOT_FOUND:errno.__dict__.get('ENOENT', None),
    LIBUSB_ERROR_BUSY:errno.__dict__.get('EBUSY', None),
    LIBUSB_ERROR_TIMEOUT:errno.__dict__.get('ETIMEDOUT', None),
    LIBUSB_ERROR_OVERFLOW:errno.__dict__.get('EOVERFLOW', None),
    LIBUSB_ERROR_PIPE:errno.__dict__.get('EPIPE', None),
    LIBUSB_ERROR_INTERRUPTED:errno.__dict__.get('EINTR', None),
    LIBUSB_ERROR_NO_MEM:errno.__dict__.get('ENOMEM', None),
    LIBUSB_ERROR_NOT_SUPPORTED:errno.__dict__.get('ENOSYS', None),
    LIBUSB_ERROR_OTHER:None
}

# Transfer status codes:
# Note that this does not indicate
# that the entire amount of requested data was transferred.
LIBUSB_TRANSFER_COMPLETED = 0
LIBUSB_TRANSFER_ERROR = 1
LIBUSB_TRANSFER_TIMED_OUT = 2
LIBUSB_TRANSFER_CANCELLED = 3
LIBUSB_TRANSFER_STALL = 4
LIBUSB_TRANSFER_NO_DEVICE = 5
LIBUSB_TRANSFER_OVERFLOW = 6

# map return codes to strings
_str_transfer_error = {
    LIBUSB_TRANSFER_COMPLETED:'Success (no error)',
    LIBUSB_TRANSFER_ERROR:'Transfer failed',
    LIBUSB_TRANSFER_TIMED_OUT:'Transfer timed out',
    LIBUSB_TRANSFER_CANCELLED:'Transfer was cancelled',
    LIBUSB_TRANSFER_STALL:'For bulk/interrupt endpoints: halt condition '\
                          'detected (endpoint stalled). For control '\
                          'endpoints: control request not supported.',
    LIBUSB_TRANSFER_NO_DEVICE:'Device was disconnected',
    LIBUSB_TRANSFER_OVERFLOW:'Device sent more data than requested'
}

# map transfer codes to errno codes
_transfer_errno = {
    LIBUSB_TRANSFER_COMPLETED:0,
    LIBUSB_TRANSFER_ERROR:errno.__dict__.get('EIO', None),
    LIBUSB_TRANSFER_TIMED_OUT:errno.__dict__.get('ETIMEDOUT', None),
    LIBUSB_TRANSFER_CANCELLED:errno.__dict__.get('EAGAIN', None),
    LIBUSB_TRANSFER_STALL:errno.__dict__.get('EIO', None),
    LIBUSB_TRANSFER_NO_DEVICE:errno.__dict__.get('ENODEV', None),
    LIBUSB_TRANSFER_OVERFLOW:errno.__dict__.get('EOVERFLOW', None)
}

def _strerror(errcode):
    try:
        return _lib.libusb_strerror(errcode).decode('utf8')
    except AttributeError:
        return _str_error_map[errcode]

# Data structures

class _libusb_endpoint_descriptor(Structure):
    _fields_ = [('bLength', c_uint8),
                ('bDescriptorType', c_uint8),
                ('bEndpointAddress', c_uint8),
                ('bmAttributes', c_uint8),
                ('wMaxPacketSize', c_uint16),
                ('bInterval', c_uint8),
                ('bRefresh', c_uint8),
                ('bSynchAddress', c_uint8),
                ('extra', POINTER(c_ubyte)),
                ('extra_length', c_int)]

class _libusb_interface_descriptor(Structure):
    _fields_ = [('bLength', c_uint8),
                ('bDescriptorType', c_uint8),
                ('bInterfaceNumber', c_uint8),
                ('bAlternateSetting', c_uint8),
                ('bNumEndpoints', c_uint8),
                ('bInterfaceClass', c_uint8),
                ('bInterfaceSubClass', c_uint8),
                ('bInterfaceProtocol', c_uint8),
                ('iInterface', c_uint8),
                ('endpoint', POINTER(_libusb_endpoint_descriptor)),
                ('extra', POINTER(c_ubyte)),
                ('extra_length', c_int)]

class _libusb_interface(Structure):
    _fields_ = [('altsetting', POINTER(_libusb_interface_descriptor)),
                ('num_altsetting', c_int)]

class _libusb_config_descriptor(Structure):
    _fields_ = [('bLength', c_uint8),
                ('bDescriptorType', c_uint8),
                ('wTotalLength', c_uint16),
                ('bNumInterfaces', c_uint8),
                ('bConfigurationValue', c_uint8),
                ('iConfiguration', c_uint8),
                ('bmAttributes', c_uint8),
                ('bMaxPower', c_uint8),
                ('interface', POINTER(_libusb_interface)),
                ('extra', POINTER(c_ubyte)),
                ('extra_length', c_int)]

class _libusb_device_descriptor(Structure):
    _fields_ = [('bLength', c_uint8),
                ('bDescriptorType', c_uint8),
                ('bcdUSB', c_uint16),
                ('bDeviceClass', c_uint8),
                ('bDeviceSubClass', c_uint8),
                ('bDeviceProtocol', c_uint8),
                ('bMaxPacketSize0', c_uint8),
                ('idVendor', c_uint16),
                ('idProduct', c_uint16),
                ('bcdDevice', c_uint16),
                ('iManufacturer', c_uint8),
                ('iProduct', c_uint8),
                ('iSerialNumber', c_uint8),
                ('bNumConfigurations', c_uint8)]


# Isochronous packet descriptor.
class _libusb_iso_packet_descriptor(Structure):
    _fields_ = [('length', c_uint),
                ('actual_length', c_uint),
                ('status', c_int)] # enum libusb_transfer_status

_libusb_device_handle = c_void_p

class _libusb_transfer(Structure):
    pass
_libusb_transfer_p = POINTER(_libusb_transfer)

_libusb_transfer_cb_fn_p = CFUNCTYPE(None, _libusb_transfer_p)

_libusb_transfer._fields_ = [('dev_handle', _libusb_device_handle),
                             ('flags', c_uint8),
                             ('endpoint', c_uint8),
                             ('type', c_uint8),
                             ('timeout', c_uint),
                             ('status', c_int), # enum libusb_transfer_status
                             ('length', c_int),
                             ('actual_length', c_int),
                             ('callback', _libusb_transfer_cb_fn_p),
                             ('user_data', py_object),
                             ('buffer', c_void_p),
                             ('num_iso_packets', c_int),
                             ('iso_packet_desc', _libusb_iso_packet_descriptor)
]

def _get_iso_packet_list(transfer):
    list_type = _libusb_iso_packet_descriptor * transfer.num_iso_packets
    return list_type.from_address(addressof(transfer.iso_packet_desc))

_lib = None
_lib_object = None

def _load_library(find_library=None):
    # Windows backend uses stdcall calling convention
    #
    # On FreeBSD 8/9, libusb 1.0 and libusb 0.1 are in the same shared
    # object libusb.so, so if we found libusb library name, we must assure
    # it is 1.0 version. We just try to get some symbol from 1.0 version
    if sys.platform == 'win32':
        win_cls = WinDLL
    else:
        win_cls = None

    return usb.libloader.load_locate_library(
                ('usb-1.0', 'libusb-1.0', 'usb'),
                'cygusb-1.0.dll', 'Libusb 1',
                win_cls=win_cls,
                find_library=find_library, check_symbols=('libusb_init',))

def _setup_prototypes(lib):
    # void libusb_set_debug (libusb_context *ctx, int level)
    lib.libusb_set_debug.argtypes = [c_void_p, c_int]

    # int libusb_init (libusb_context **context)
    lib.libusb_init.argtypes = [POINTER(c_void_p)]

    # void libusb_exit (struct libusb_context *ctx)
    lib.libusb_exit.argtypes = [c_void_p]

    # ssize_t libusb_get_device_list (libusb_context *ctx,
    #                                 libusb_device ***list)
    lib.libusb_get_device_list.argtypes = [
            c_void_p,
            POINTER(POINTER(c_void_p))
        ]

    # libusb_device *libusb_get_parent (libusb_device *dev)
    lib.libusb_get_parent.argtypes = [c_void_p]
    lib.libusb_get_parent.restype = c_void_p

    # void libusb_free_device_list (libusb_device **list,
    #                               int unref_devices)
    lib.libusb_free_device_list.argtypes = [
            POINTER(c_void_p),
            c_int
        ]

    # libusb_device *libusb_ref_device (libusb_device *dev)
    lib.libusb_ref_device.argtypes = [c_void_p]
    lib.libusb_ref_device.restype = c_void_p

    # void libusb_unref_device(libusb_device *dev)
    lib.libusb_unref_device.argtypes = [c_void_p]

    # int libusb_open(libusb_device *dev, libusb_device_handle **handle)
    lib.libusb_open.argtypes = [c_void_p, POINTER(_libusb_device_handle)]

    # void libusb_close(libusb_device_handle *dev_handle)
    lib.libusb_close.argtypes = [_libusb_device_handle]

    # int libusb_set_configuration(libusb_device_handle *dev,
    #                              int configuration)
    lib.libusb_set_configuration.argtypes = [_libusb_device_handle, c_int]

    # int libusb_get_configuration(libusb_device_handle *dev,
    #                              int *config)
    lib.libusb_get_configuration.argtypes = [_libusb_device_handle, POINTER(c_int)]

    # int libusb_claim_interface(libusb_device_handle *dev,
    #                               int interface_number)
    lib.libusb_claim_interface.argtypes = [_libusb_device_handle, c_int]

    # int libusb_release_interface(libusb_device_handle *dev,
    #                              int interface_number)
    lib.libusb_release_interface.argtypes = [_libusb_device_handle, c_int]

    # int libusb_set_interface_alt_setting(libusb_device_handle *dev,
    #                                      int interface_number,
    #                                      int alternate_setting)
    lib.libusb_set_interface_alt_setting.argtypes = [
            _libusb_device_handle,
            c_int,
            c_int
        ]

    # int libusb_reset_device (libusb_device_handle *dev)
    lib.libusb_reset_device.argtypes = [_libusb_device_handle]

    # int libusb_kernel_driver_active(libusb_device_handle *dev,
    #                                 int interface)
    lib.libusb_kernel_driver_active.argtypes = [
            _libusb_device_handle,
            c_int
        ]

    # int libusb_detach_kernel_driver(libusb_device_handle *dev,
    #                                 int interface)
    lib.libusb_detach_kernel_driver.argtypes = [
            _libusb_device_handle,
            c_int
        ]

    # int libusb_attach_kernel_driver(libusb_device_handle *dev,
    #                                 int interface)
    lib.libusb_attach_kernel_driver.argtypes = [
            _libusb_device_handle,
            c_int
        ]

    # int libusb_get_device_descriptor(
    #                   libusb_device *dev,
    #                   struct libusb_device_descriptor *desc
    #               )
    lib.libusb_get_device_descriptor.argtypes = [
            c_void_p,
            POINTER(_libusb_device_descriptor)
        ]

    # int libusb_get_config_descriptor(
    #           libusb_device *dev,
    #           uint8_t config_index,
    #           struct libusb_config_descriptor **config
    #       )
    lib.libusb_get_config_descriptor.argtypes = [
            c_void_p,
            c_uint8,
            POINTER(POINTER(_libusb_config_descriptor))
        ]

    # void  libusb_free_config_descriptor(
    #           struct libusb_config_descriptor *config
    #   )
    lib.libusb_free_config_descriptor.argtypes = [
            POINTER(_libusb_config_descriptor)
        ]

    # int libusb_get_string_descriptor_ascii(libusb_device_handle *dev,
    #                                         uint8_t desc_index,
    #                                         unsigned char *data,
    #                                         int length)
    lib.libusb_get_string_descriptor_ascii.argtypes = [
            _libusb_device_handle,
            c_uint8,
            POINTER(c_ubyte),
            c_int
        ]

    # int libusb_control_transfer(libusb_device_handle *dev_handle,
    #                             uint8_t bmRequestType,
    #                             uint8_t bRequest,
    #                             uint16_t wValue,
    #                             uint16_t wIndex,
    #                             unsigned char *data,
    #                             uint16_t wLength,
    #                             unsigned int timeout)
    lib.libusb_control_transfer.argtypes = [
            _libusb_device_handle,
            c_uint8,
            c_uint8,
            c_uint16,
            c_uint16,
            POINTER(c_ubyte),
            c_uint16,
            c_uint
        ]

    #int libusb_bulk_transfer(
    #           struct libusb_device_handle *dev_handle,
    #           unsigned char endpoint,
    #           unsigned char *data,
    #           int length,
    #           int *transferred,
    #           unsigned int timeout
    #       )
    lib.libusb_bulk_transfer.argtypes = [
                _libusb_device_handle,
                c_ubyte,
                POINTER(c_ubyte),
                c_int,
                POINTER(c_int),
                c_uint
            ]

    # int libusb_interrupt_transfer(
    #               libusb_device_handle *dev_handle,
    #               unsigned char endpoint,
    #               unsigned char *data,
    #               int length,
    #               int *actual_length,
    #               unsigned int timeout
    #           );
    lib.libusb_interrupt_transfer.argtypes = [
                    _libusb_device_handle,
                    c_ubyte,
                    POINTER(c_ubyte),
                    c_int,
                    POINTER(c_int),
                    c_uint
                ]

    # libusb_transfer* libusb_alloc_transfer(int iso_packets);
    lib.libusb_alloc_transfer.argtypes = [c_int]
    lib.libusb_alloc_transfer.restype = POINTER(_libusb_transfer)

    # void libusb_free_transfer(struct libusb_transfer *transfer)
    lib.libusb_free_transfer.argtypes = [POINTER(_libusb_transfer)]

    # int libusb_submit_transfer(struct libusb_transfer *transfer);
    lib.libusb_submit_transfer.argtypes = [POINTER(_libusb_transfer)]

    if hasattr(lib, 'libusb_strerror'):
        # const char *libusb_strerror(enum libusb_error errcode)
        lib.libusb_strerror.argtypes = [c_uint]
        lib.libusb_strerror.restype = c_char_p

    # int libusb_clear_halt(libusb_device_handle *dev, unsigned char endpoint)
    lib.libusb_clear_halt.argtypes = [_libusb_device_handle, c_ubyte]

    # void libusb_set_iso_packet_lengths(
    #               libusb_transfer* transfer,
    #               unsigned int length
    #           );
    def libusb_set_iso_packet_lengths(transfer_p, length):
        r"""This function is inline in the libusb.h file, so we must implement
            it.

        lib.libusb_set_iso_packet_lengths.argtypes = [
                        POINTER(_libusb_transfer),
                        c_int
                    ]
        """
        transfer = transfer_p.contents
        for iso_packet_desc in _get_iso_packet_list(transfer):
            iso_packet_desc.length = length
    lib.libusb_set_iso_packet_lengths = libusb_set_iso_packet_lengths

    #int libusb_get_max_iso_packet_size(libusb_device* dev,
    #                                   unsigned char endpoint);
    lib.libusb_get_max_iso_packet_size.argtypes = [c_void_p,
                                                   c_ubyte]

    # void libusb_fill_iso_transfer(
    #               struct libusb_transfer* transfer,
    #               libusb_device_handle*  dev_handle,
    #               unsigned char endpoint,
    #               unsigned char* buffer,
    #               int length,
    #               int num_iso_packets,
    #               libusb_transfer_cb_fn   callback,
    #               void * user_data,
    #               unsigned int timeout
    #           );
    def libusb_fill_iso_transfer(_libusb_transfer_p, dev_handle, endpoint, buffer, length,
                                 num_iso_packets, callback, user_data, timeout):
        r"""This function is inline in the libusb.h file, so we must implement
            it.

        lib.libusb_fill_iso_transfer.argtypes = [
                       _libusb_transfer,
                       _libusb_device_handle,
                       c_ubyte,
                       POINTER(c_ubyte),
                       c_int,
                       c_int,
                       _libusb_transfer_cb_fn_p,
                       c_void_p,
                       c_uint
                   ]
        """
        transfer = _libusb_transfer_p.contents
        transfer.dev_handle = dev_handle
        transfer.endpoint = endpoint
        transfer.type = _LIBUSB_TRANSFER_TYPE_ISOCHRONOUS
        transfer.timeout = timeout
        transfer.buffer = cast(buffer, c_void_p)
        transfer.length = length
        transfer.num_iso_packets = num_iso_packets
        transfer.user_data = user_data
        transfer.callback = callback
    lib.libusb_fill_iso_transfer = libusb_fill_iso_transfer

    # uint8_t libusb_get_bus_number(libusb_device *dev)
    lib.libusb_get_bus_number.argtypes = [c_void_p]
    lib.libusb_get_bus_number.restype = c_uint8

    # uint8_t libusb_get_device_address(libusb_device *dev)
    lib.libusb_get_device_address.argtypes = [c_void_p]
    lib.libusb_get_device_address.restype = c_uint8

    try:
        # uint8_t libusb_get_device_speed(libusb_device *dev)
        lib.libusb_get_device_speed.argtypes = [c_void_p]
        lib.libusb_get_device_speed.restype = c_uint8
    except AttributeError:
        pass

    try:
        # uint8_t libusb_get_port_number(libusb_device *dev)
        lib.libusb_get_port_number.argtypes = [c_void_p]
        lib.libusb_get_port_number.restype = c_uint8
    except AttributeError:
        pass

    try:
        # int libusb_get_port_numbers(libusb_device *dev,
        #                             uint8_t* port_numbers,
        #                             int port_numbers_len)
        lib.libusb_get_port_numbers.argtypes = [
                c_void_p,
                POINTER(c_uint8),
                c_int
            ]
        lib.libusb_get_port_numbers.restype = c_int
    except AttributeError:
        pass

    #int libusb_handle_events(libusb_context *ctx);
    lib.libusb_handle_events.argtypes = [c_void_p]

# check a libusb function call
def _check(ret):
    if hasattr(ret, 'value'):
        ret = ret.value

    if ret < 0:
        if ret == LIBUSB_ERROR_NOT_SUPPORTED:
            raise NotImplementedError(_strerror(ret))
        elif ret == LIBUSB_ERROR_TIMEOUT:
            raise USBTimeoutError(_strerror(ret), ret, _libusb_errno[ret])
        else:
            raise USBError(_strerror(ret), ret, _libusb_errno[ret])

    return ret

# wrap a device
class _Device(_objfinalizer.AutoFinalizedObject):
    def __init__(self, devid):
        self.devid = _lib.libusb_ref_device(devid)
    def _finalize_object(self):
        _lib.libusb_unref_device(self.devid)

# wrap a descriptor and keep a reference to another object
# Thanks to Thomas Reitmayr.
class _WrapDescriptor(object):
    def __init__(self, desc, obj = None):
        self.obj = obj
        self.desc = desc
    def __getattr__(self, name):
        return getattr(self.desc, name)

# wrap a configuration descriptor
class _ConfigDescriptor(_objfinalizer.AutoFinalizedObject):
    def __init__(self, desc):
        self.desc = desc
    def _finalize_object(self):
        _lib.libusb_free_config_descriptor(self.desc)
    def __getattr__(self, name):
        return getattr(self.desc.contents, name)


# iterator for libusb devices
class _DevIterator(_objfinalizer.AutoFinalizedObject):
    def __init__(self, ctx):
        self.dev_list = POINTER(c_void_p)()
        self.num_devs = _check(_lib.libusb_get_device_list(
                                    ctx,
                                    byref(self.dev_list))
                                )
    def __iter__(self):
        for i in range(self.num_devs):
            yield _Device(self.dev_list[i])
    def _finalize_object(self):
        _lib.libusb_free_device_list(self.dev_list, 1)

class _DeviceHandle(object):
    def __init__(self, dev):
        self.handle = _libusb_device_handle()
        self.devid = dev.devid
        _check(_lib.libusb_open(self.devid, byref(self.handle)))

class _IsoTransferHandler(_objfinalizer.AutoFinalizedObject):
    def __init__(self, dev_handle, ep, buff, timeout):
        address, length = buff.buffer_info()

        packet_length = _lib.libusb_get_max_iso_packet_size(dev_handle.devid, ep)
        packet_count = int(math.ceil(float(length) / packet_length))

        self.transfer = _lib.libusb_alloc_transfer(packet_count)

        _lib.libusb_fill_iso_transfer(self.transfer,
                                      dev_handle.handle,
                                      ep,
                                      cast(address, POINTER(c_ubyte)),
                                      length,
                                      packet_count,
                                      _libusb_transfer_cb_fn_p(self.__callback),
                                      None,
                                      timeout)

        self.__set_packets_length(length, packet_length)

    def _finalize_object(self):
        _lib.libusb_free_transfer(self.transfer)

    def submit(self, ctx = None):
        self.__callback_done = 0
        _check(_lib.libusb_submit_transfer(self.transfer))

        while not self.__callback_done:
            _check(_lib.libusb_handle_events(ctx))

        status = int(self.transfer.contents.status)
        if status != LIBUSB_TRANSFER_COMPLETED:
            raise usb.USBError(_str_transfer_error[status],
                               status,
                               _transfer_errno[status])

        return self.__compute_size_transf_data()

    def __compute_size_transf_data(self):
        return sum([t.actual_length for t in
                    _get_iso_packet_list(self.transfer.contents)])

    def __set_packets_length(self, n, packet_length):
        _lib.libusb_set_iso_packet_lengths(self.transfer, packet_length)
        r = n % packet_length
        if r:
            iso_packets = _get_iso_packet_list(self.transfer.contents)
            # When the device is disconnected, this list may
            # return with length 0
            if len(iso_packets):
                iso_packets[-1].length = r

    def __callback(self, transfer):
        self.__callback_done = 1

# implementation of libusb 1.0 backend
class _LibUSB(usb.backend.IBackend):
    @methodtrace(_logger)
    def __init__(self, lib):
        usb.backend.IBackend.__init__(self)
        self.lib = lib
        self.ctx = c_void_p()
        _check(self.lib.libusb_init(byref(self.ctx)))

    @methodtrace(_logger)
    def _finalize_object(self):
        if self.ctx:
            self.lib.libusb_exit(self.ctx)


    @methodtrace(_logger)
    def enumerate_devices(self):
        return _DevIterator(self.ctx)

    @methodtrace(_logger)
    def get_parent(self, dev):
        _parent = self.lib.libusb_get_parent(dev.devid)
        if _parent is None:
            return None
        else:
            return _Device(_parent)

    @methodtrace(_logger)
    def get_device_descriptor(self, dev):
        dev_desc = _libusb_device_descriptor()
        _check(self.lib.libusb_get_device_descriptor(dev.devid, byref(dev_desc)))
        dev_desc.bus = self.lib.libusb_get_bus_number(dev.devid)
        dev_desc.address = self.lib.libusb_get_device_address(dev.devid)

        # Only available in newer versions of libusb
        try:
            dev_desc.speed = self.lib.libusb_get_device_speed(dev.devid)
        except AttributeError:
            dev_desc.speed = None

        # Only available in newer versions of libusb
        try:
            dev_desc.port_number = self.lib.libusb_get_port_number(dev.devid)
        except AttributeError:
            dev_desc.port_number = None

        # Only available in newer versions of libusb
        try:
            buff = (c_uint8 * 7)()  # USB 3.0 maximum depth is 7
            written = dev_desc.port_numbers = self.lib.libusb_get_port_numbers(
                    dev.devid, buff, len(buff))
            if written > 0:
                dev_desc.port_numbers = tuple(buff[:written])
            else:
                dev_desc.port_numbers = None
        except AttributeError:
            dev_desc.port_numbers = None

        return dev_desc

    @methodtrace(_logger)
    def get_configuration_descriptor(self, dev, config):
        cfg = POINTER(_libusb_config_descriptor)()
        _check(self.lib.libusb_get_config_descriptor(
                dev.devid,
                config, byref(cfg)))
        config_desc = _ConfigDescriptor(cfg)
        config_desc.extra_descriptors = (
                config_desc.extra[:config_desc.extra_length])
        return config_desc

    @methodtrace(_logger)
    def get_interface_descriptor(self, dev, intf, alt, config):
        cfg = self.get_configuration_descriptor(dev, config)
        if intf >= cfg.bNumInterfaces:
            raise IndexError('Invalid interface index ' + str(intf))
        i = cfg.interface[intf]
        if alt >= i.num_altsetting:
            raise IndexError('Invalid alternate setting index ' + str(alt))
        intf_desc = i.altsetting[alt]
        intf_desc.extra_descriptors = intf_desc.extra[:intf_desc.extra_length]
        return _WrapDescriptor(intf_desc, cfg)

    @methodtrace(_logger)
    def get_endpoint_descriptor(self, dev, ep, intf, alt, config):
        i = self.get_interface_descriptor(dev, intf, alt, config)
        if ep > i.bNumEndpoints:
            raise IndexError('Invalid endpoint index ' + str(ep))
        ep_desc = i.endpoint[ep]
        ep_desc.extra_descriptors = ep_desc.extra[:ep_desc.extra_length]
        return _WrapDescriptor(ep_desc, i)

    @methodtrace(_logger)
    def open_device(self, dev):
        return _DeviceHandle(dev)

    @methodtrace(_logger)
    def close_device(self, dev_handle):
        self.lib.libusb_close(dev_handle.handle)

    @methodtrace(_logger)
    def set_configuration(self, dev_handle, config_value):
        _check(self.lib.libusb_set_configuration(dev_handle.handle, config_value))

    @methodtrace(_logger)
    def get_configuration(self, dev_handle):
        config = c_int()
        _check(self.lib.libusb_get_configuration(dev_handle.handle, byref(config)))
        return config.value

    @methodtrace(_logger)
    def set_interface_altsetting(self, dev_handle, intf, altsetting):
        _check(self.lib.libusb_set_interface_alt_setting(
                                dev_handle.handle,
                                intf,
                                altsetting))

    @methodtrace(_logger)
    def claim_interface(self, dev_handle, intf):
        _check(self.lib.libusb_claim_interface(dev_handle.handle, intf))

    @methodtrace(_logger)
    def release_interface(self, dev_handle, intf):
        _check(self.lib.libusb_release_interface(dev_handle.handle, intf))

    @methodtrace(_logger)
    def bulk_write(self, dev_handle, ep, intf, data, timeout):
        return self.__write(self.lib.libusb_bulk_transfer,
                            dev_handle,
                            ep,
                            intf,
                            data,
                            timeout)

    @methodtrace(_logger)
    def bulk_read(self, dev_handle, ep, intf, buff, timeout):
        return self.__read(self.lib.libusb_bulk_transfer,
                           dev_handle,
                           ep,
                           intf,
                           buff,
                           timeout)

    @methodtrace(_logger)
    def intr_write(self, dev_handle, ep, intf, data, timeout):
        return self.__write(self.lib.libusb_interrupt_transfer,
                            dev_handle,
                            ep,
                            intf,
                            data,
                            timeout)

    @methodtrace(_logger)
    def intr_read(self, dev_handle, ep, intf, buff, timeout):
        return self.__read(self.lib.libusb_interrupt_transfer,
                           dev_handle,
                           ep,
                           intf,
                           buff,
                           timeout)

    @methodtrace(_logger)
    def iso_write(self, dev_handle, ep, intf, data, timeout):
        handler = _IsoTransferHandler(dev_handle, ep, data, timeout)
        return handler.submit(self.ctx)

    @methodtrace(_logger)
    def iso_read(self, dev_handle, ep, intf, buff, timeout):
        handler = _IsoTransferHandler(dev_handle, ep, buff, timeout)
        return handler.submit(self.ctx)

    @methodtrace(_logger)
    def ctrl_transfer(self,
                      dev_handle,
                      bmRequestType,
                      bRequest,
                      wValue,
                      wIndex,
                      data,
                      timeout):
        addr, length = data.buffer_info()
        length *= data.itemsize

        ret = _check(self.lib.libusb_control_transfer(
                                        dev_handle.handle,
                                        bmRequestType,
                                        bRequest,
                                        wValue,
                                        wIndex,
                                        cast(addr, POINTER(c_ubyte)),
                                        length,
                                        timeout))

        return ret

    @methodtrace(_logger)
    def clear_halt(self, dev_handle, ep):
        _check(self.lib.libusb_clear_halt(dev_handle.handle, ep))

    @methodtrace(_logger)
    def reset_device(self, dev_handle):
        _check(self.lib.libusb_reset_device(dev_handle.handle))

    @methodtrace(_logger)
    def is_kernel_driver_active(self, dev_handle, intf):
        return bool(_check(self.lib.libusb_kernel_driver_active(dev_handle.handle,
                        intf)))

    @methodtrace(_logger)
    def detach_kernel_driver(self, dev_handle, intf):
        _check(self.lib.libusb_detach_kernel_driver(dev_handle.handle, intf))

    @methodtrace(_logger)
    def attach_kernel_driver(self, dev_handle, intf):
        _check(self.lib.libusb_attach_kernel_driver(dev_handle.handle, intf))

    def __write(self, fn, dev_handle, ep, intf, data, timeout):
        address, length = data.buffer_info()
        length *= data.itemsize
        transferred = c_int()
        retval = fn(dev_handle.handle,
                  ep,
                  cast(address, POINTER(c_ubyte)),
                  length,
                  byref(transferred),
                  timeout)
        # do not assume LIBUSB_ERROR_TIMEOUT means no I/O.
        if not (transferred.value and retval == LIBUSB_ERROR_TIMEOUT):
            _check(retval)

        return transferred.value

    def __read(self, fn, dev_handle, ep, intf, buff, timeout):
        address, length = buff.buffer_info()
        length *= buff.itemsize
        transferred = c_int()
        retval = fn(dev_handle.handle,
                  ep,
                  cast(address, POINTER(c_ubyte)),
                  length,
                  byref(transferred),
                  timeout)
        # do not assume LIBUSB_ERROR_TIMEOUT means no I/O.
        if not (transferred.value and retval == LIBUSB_ERROR_TIMEOUT):
            _check(retval)
        return transferred.value

def get_backend(find_library=None):
    global _lib, _lib_object
    try:
        if _lib is None:
            _lib = _load_library(find_library=find_library)
            _setup_prototypes(_lib)
        if _lib_object is None:
            _lib_object = _LibUSB(_lib)
        return _lib_object
    except usb.libloader.LibraryException:
        # exception already logged (if any)
        _logger.error('Error loading libusb 1.0 backend', exc_info=False)
        return None
    except Exception:
        _logger.error('Error loading libusb 1.0 backend', exc_info=True)
        return None



"""
UNFINISHED
For the ``future`` package.

Adds this import line:

    from __future__ import division

at the top so the code runs identically on Py3 and Py2.6/2.7
"""

from libpasteurize.fixes.fix_division import FixDivision




import bustoken
import json
import sysactions
from msgbus import TK_SYS_ACK
from typing import Dict

from model.customexception import MessageBusError
from model.productmanager import PartWithDependencies

TK_SLCTRL_PMGR_GET_PARTS_WITH_DEPENDENCIES = bustoken.create_token(
    priority=bustoken.MSGPRT_LOW,
    message_grourp="9",
    number="1",
)


class ProductManager(object):
    _parts_with_dependencies_cache = None

    @staticmethod
    def get_parts_with_dependencies():
        # type: (...) -> Dict[int, PartWithDependencies]

        if ProductManager._parts_with_dependencies_cache is not None:
            return ProductManager._parts_with_dependencies_cache

        msg = sysactions.send_message("ORDERMGR", TK_SLCTRL_PMGR_GET_PARTS_WITH_DEPENDENCIES)

        if msg.token != TK_SYS_ACK:
            raise MessageBusError(
                component="ORDERMGR",
                token=msg.token,
                data=msg.data
            )

        json_dict = json.loads(msg.data.decode("utf-8"))
        ret = {}
        for part_code, part in json_dict.items():
            part_code = int(part_code)
            ret[part_code] = PartWithDependencies(
                part_code=part_code,
                name=part["name"],
                dependencies=part["dependencies"],
            )

        ProductManager._parts_with_dependencies_cache = ret
        return ret



data = (
'mya',    # 0x00
'myag',    # 0x01
'myagg',    # 0x02
'myags',    # 0x03
'myan',    # 0x04
'myanj',    # 0x05
'myanh',    # 0x06
'myad',    # 0x07
'myal',    # 0x08
'myalg',    # 0x09
'myalm',    # 0x0a
'myalb',    # 0x0b
'myals',    # 0x0c
'myalt',    # 0x0d
'myalp',    # 0x0e
'myalh',    # 0x0f
'myam',    # 0x10
'myab',    # 0x11
'myabs',    # 0x12
'myas',    # 0x13
'myass',    # 0x14
'myang',    # 0x15
'myaj',    # 0x16
'myac',    # 0x17
'myak',    # 0x18
'myat',    # 0x19
'myap',    # 0x1a
'myah',    # 0x1b
'myae',    # 0x1c
'myaeg',    # 0x1d
'myaegg',    # 0x1e
'myaegs',    # 0x1f
'myaen',    # 0x20
'myaenj',    # 0x21
'myaenh',    # 0x22
'myaed',    # 0x23
'myael',    # 0x24
'myaelg',    # 0x25
'myaelm',    # 0x26
'myaelb',    # 0x27
'myaels',    # 0x28
'myaelt',    # 0x29
'myaelp',    # 0x2a
'myaelh',    # 0x2b
'myaem',    # 0x2c
'myaeb',    # 0x2d
'myaebs',    # 0x2e
'myaes',    # 0x2f
'myaess',    # 0x30
'myaeng',    # 0x31
'myaej',    # 0x32
'myaec',    # 0x33
'myaek',    # 0x34
'myaet',    # 0x35
'myaep',    # 0x36
'myaeh',    # 0x37
'meo',    # 0x38
'meog',    # 0x39
'meogg',    # 0x3a
'meogs',    # 0x3b
'meon',    # 0x3c
'meonj',    # 0x3d
'meonh',    # 0x3e
'meod',    # 0x3f
'meol',    # 0x40
'meolg',    # 0x41
'meolm',    # 0x42
'meolb',    # 0x43
'meols',    # 0x44
'meolt',    # 0x45
'meolp',    # 0x46
'meolh',    # 0x47
'meom',    # 0x48
'meob',    # 0x49
'meobs',    # 0x4a
'meos',    # 0x4b
'meoss',    # 0x4c
'meong',    # 0x4d
'meoj',    # 0x4e
'meoc',    # 0x4f
'meok',    # 0x50
'meot',    # 0x51
'meop',    # 0x52
'meoh',    # 0x53
'me',    # 0x54
'meg',    # 0x55
'megg',    # 0x56
'megs',    # 0x57
'men',    # 0x58
'menj',    # 0x59
'menh',    # 0x5a
'med',    # 0x5b
'mel',    # 0x5c
'melg',    # 0x5d
'melm',    # 0x5e
'melb',    # 0x5f
'mels',    # 0x60
'melt',    # 0x61
'melp',    # 0x62
'melh',    # 0x63
'mem',    # 0x64
'meb',    # 0x65
'mebs',    # 0x66
'mes',    # 0x67
'mess',    # 0x68
'meng',    # 0x69
'mej',    # 0x6a
'mec',    # 0x6b
'mek',    # 0x6c
'met',    # 0x6d
'mep',    # 0x6e
'meh',    # 0x6f
'myeo',    # 0x70
'myeog',    # 0x71
'myeogg',    # 0x72
'myeogs',    # 0x73
'myeon',    # 0x74
'myeonj',    # 0x75
'myeonh',    # 0x76
'myeod',    # 0x77
'myeol',    # 0x78
'myeolg',    # 0x79
'myeolm',    # 0x7a
'myeolb',    # 0x7b
'myeols',    # 0x7c
'myeolt',    # 0x7d
'myeolp',    # 0x7e
'myeolh',    # 0x7f
'myeom',    # 0x80
'myeob',    # 0x81
'myeobs',    # 0x82
'myeos',    # 0x83
'myeoss',    # 0x84
'myeong',    # 0x85
'myeoj',    # 0x86
'myeoc',    # 0x87
'myeok',    # 0x88
'myeot',    # 0x89
'myeop',    # 0x8a
'myeoh',    # 0x8b
'mye',    # 0x8c
'myeg',    # 0x8d
'myegg',    # 0x8e
'myegs',    # 0x8f
'myen',    # 0x90
'myenj',    # 0x91
'myenh',    # 0x92
'myed',    # 0x93
'myel',    # 0x94
'myelg',    # 0x95
'myelm',    # 0x96
'myelb',    # 0x97
'myels',    # 0x98
'myelt',    # 0x99
'myelp',    # 0x9a
'myelh',    # 0x9b
'myem',    # 0x9c
'myeb',    # 0x9d
'myebs',    # 0x9e
'myes',    # 0x9f
'myess',    # 0xa0
'myeng',    # 0xa1
'myej',    # 0xa2
'myec',    # 0xa3
'myek',    # 0xa4
'myet',    # 0xa5
'myep',    # 0xa6
'myeh',    # 0xa7
'mo',    # 0xa8
'mog',    # 0xa9
'mogg',    # 0xaa
'mogs',    # 0xab
'mon',    # 0xac
'monj',    # 0xad
'monh',    # 0xae
'mod',    # 0xaf
'mol',    # 0xb0
'molg',    # 0xb1
'molm',    # 0xb2
'molb',    # 0xb3
'mols',    # 0xb4
'molt',    # 0xb5
'molp',    # 0xb6
'molh',    # 0xb7
'mom',    # 0xb8
'mob',    # 0xb9
'mobs',    # 0xba
'mos',    # 0xbb
'moss',    # 0xbc
'mong',    # 0xbd
'moj',    # 0xbe
'moc',    # 0xbf
'mok',    # 0xc0
'mot',    # 0xc1
'mop',    # 0xc2
'moh',    # 0xc3
'mwa',    # 0xc4
'mwag',    # 0xc5
'mwagg',    # 0xc6
'mwags',    # 0xc7
'mwan',    # 0xc8
'mwanj',    # 0xc9
'mwanh',    # 0xca
'mwad',    # 0xcb
'mwal',    # 0xcc
'mwalg',    # 0xcd
'mwalm',    # 0xce
'mwalb',    # 0xcf
'mwals',    # 0xd0
'mwalt',    # 0xd1
'mwalp',    # 0xd2
'mwalh',    # 0xd3
'mwam',    # 0xd4
'mwab',    # 0xd5
'mwabs',    # 0xd6
'mwas',    # 0xd7
'mwass',    # 0xd8
'mwang',    # 0xd9
'mwaj',    # 0xda
'mwac',    # 0xdb
'mwak',    # 0xdc
'mwat',    # 0xdd
'mwap',    # 0xde
'mwah',    # 0xdf
'mwae',    # 0xe0
'mwaeg',    # 0xe1
'mwaegg',    # 0xe2
'mwaegs',    # 0xe3
'mwaen',    # 0xe4
'mwaenj',    # 0xe5
'mwaenh',    # 0xe6
'mwaed',    # 0xe7
'mwael',    # 0xe8
'mwaelg',    # 0xe9
'mwaelm',    # 0xea
'mwaelb',    # 0xeb
'mwaels',    # 0xec
'mwaelt',    # 0xed
'mwaelp',    # 0xee
'mwaelh',    # 0xef
'mwaem',    # 0xf0
'mwaeb',    # 0xf1
'mwaebs',    # 0xf2
'mwaes',    # 0xf3
'mwaess',    # 0xf4
'mwaeng',    # 0xf5
'mwaej',    # 0xf6
'mwaec',    # 0xf7
'mwaek',    # 0xf8
'mwaet',    # 0xf9
'mwaep',    # 0xfa
'mwaeh',    # 0xfb
'moe',    # 0xfc
'moeg',    # 0xfd
'moegg',    # 0xfe
'moegs',    # 0xff
)



from ._POSAlertNotificationRequest import POSAlertNotificationRequest
from ._BOHOrderPictureRequest import BOHOrderPictureRequest
from ._Configs import Configs
from services.model.exceptions import (
    BackupAlertsDatabaseNoSourceException,
    ForbiddenError,
    InUseError,
    NoResponseError,
    NotFound,
    NotFoundError,
    NotAvailableError,
    ProcessorError,
    RestoreAlertsDatabaseNoSourceException,
)
from ._ConfiguredAlert import ConfiguredAlert
from ._DiscountAppliedRequest import DiscountAppliedRequest
from ._Event import Event
from ._ListenedEvents import (
    ListenedEvents,
    all_events,
)
from ._ListenedTokens import ListenedTokens
from ._OrderCanceledRequest import OrderCanceledRequest
from ._AlertsConfigurations import AlertsConfigurations





# -*- coding: utf-8 -*-

import json

from bustoken import (
    TK_SLCTRL_PMGR_RUPTURE_UPDATE_DAILY_DISHES,
    TK_SLCTRL_PMGR_RUPTURE_UPDATE_ITEMS,
)
from msgbus import (
    FM_PARAM,
    TK_SYS_NAK,
)
from sysactions import (
    StopAction,
    action,
    show_messagebox,
)

from model.customexception import UpdateRuptureException
from .rupture_common import get_rupture_user_id_with_validation
from .. import (
    logger,
    mb_context,
)


@action
def save_daily_dishes(
    pos_id,                 # type: str
    removed_dishes_json,    # type: str
    added_dishes_json,      # type: str
):
    # type: (...) -> None

    try:
        removed_dishes = json.loads(removed_dishes_json)
        added_dishes = json.loads(added_dishes_json)

        ruptured = []
        added = []
        for added_dish in added_dishes:
            if added_dish["product_quantity"] == 0:
                ruptured.append(added_dish["product_code"])
            else:
                added.append(added_dish)

        user_id = get_rupture_user_id_with_validation(pos_id=pos_id)

        if len(added) > 0 or len(removed_dishes) > 0:
            dishes_changes_json = json.dumps(
                {
                    'unset': removed_dishes,
                    'set': added
                }
            )

            data = '\0'.join((dishes_changes_json, user_id)).encode("utf-8")
            msg = mb_context.MB_EasySendMessage(
                dest_name="ORDERMGR",
                token=TK_SLCTRL_PMGR_RUPTURE_UPDATE_DAILY_DISHES,
                format=FM_PARAM,
                data=data
            )
            if msg.token == TK_SYS_NAK:
                raise UpdateRuptureException(msg.data)

        if len(ruptured) > 0:
            rupture_status_json = json.dumps(
                {
                    "user_id": user_id,
                    "enabled": [],
                    "disabled": ruptured,
                }
            )
            msg = mb_context.MB_EasySendMessage(
                dest_name="ORDERMGR",
                token=TK_SLCTRL_PMGR_RUPTURE_UPDATE_ITEMS,
                format=FM_PARAM,
                data=rupture_status_json.encode("utf-8"))
            if msg.token == TK_SYS_NAK:
                raise UpdateRuptureException(msg.data)

        show_messagebox(
            posid=pos_id,
            message="$DAILY_DISHES_UPDATE_SUCCESS"
        )
    except StopAction:
        pass
    except Exception as e:
        logger.exception('[save_daily_dishes] Exception {}'.format(e))
        show_messagebox(
            posid=pos_id,
            message="$DAILY_DISHES_UPDATE_ERROR",
            buttons="$EXIT",
        )



# cn.py - functions for handling Japanese Corporate Number (CN)
# coding: utf-8
#
# Copyright (C) 2019 Alan Hettinger
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""CN (法人番号, hōjin bangō, Japanese Corporate Number).

The Corporate Number is assigned by the Japanese National Tax Agency to
identify government organs, public entities, registered corporations and
other organisations. The number consists of 13 digits where the first digit
is a non-zero check digit.

More information:

 * https://en.wikipedia.org/wiki/Corporate_Number
 * https://www.nta.go.jp/taxes/tetsuzuki/mynumberinfo/houjinbangou/

>>> validate('5-8356-7825-6246')
'5835678256246'
>>> validate('2-8356-7825-6246')
Traceback (most recent call last):
  ...
InvalidChecksum: ...
>>> format('5835678256246')
'5-8356-7825-6246'
"""

from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, '- ').strip()


def calc_check_digit(number):
    """Calculate the check digit. The number passed should not have
    the check digit included."""
    weights = (1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2)
    s = sum(w * int(n) for w, n in zip(weights, reversed(number))) % 9
    return str(9 - s)


def validate(number):
    """Check if the number is valid. This checks the length and check
    digit."""
    number = compact(number)
    if len(number) != 13:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    if calc_check_digit(number[1:]) != number[0]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid CN."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return '-'.join(
        (number[0], number[1:5], number[5:9], number[9:13]))



#! python
#
# Backend for .NET/Mono (IronPython), .NET >= 2
#
# This file is part of pySerial. https://github.com/pyserial/pyserial
# (C) 2008-2015 Chris Liechti <cliechti@gmx.net>
#
# SPDX-License-Identifier:    BSD-3-Clause

from __future__ import absolute_import

import System
import System.IO.Ports
from serial.serialutil import *

# must invoke function with byte array, make a helper to convert strings
# to byte arrays
sab = System.Array[System.Byte]


def as_byte_array(string):
    return sab([ord(x) for x in string])  # XXX will require adaption when run with a 3.x compatible IronPython


class Serial(SerialBase):
    """Serial port implementation for .NET/Mono."""

    BAUDRATES = (50, 75, 110, 134, 150, 200, 300, 600, 1200, 1800, 2400, 4800,
                 9600, 19200, 38400, 57600, 115200)

    def open(self):
        """\
        Open port with current settings. This may throw a SerialException
        if the port cannot be opened.
        """
        if self._port is None:
            raise SerialException("Port must be configured before it can be used.")
        if self.is_open:
            raise SerialException("Port is already open.")
        try:
            self._port_handle = System.IO.Ports.SerialPort(self.portstr)
        except Exception as msg:
            self._port_handle = None
            raise SerialException("could not open port %s: %s" % (self.portstr, msg))

        # if RTS and/or DTR are not set before open, they default to True
        if self._rts_state is None:
            self._rts_state = True
        if self._dtr_state is None:
            self._dtr_state = True

        self._reconfigure_port()
        self._port_handle.Open()
        self.is_open = True
        if not self._dsrdtr:
            self._update_dtr_state()
        if not self._rtscts:
            self._update_rts_state()
        self.reset_input_buffer()

    def _reconfigure_port(self):
        """Set communication parameters on opened port."""
        if not self._port_handle:
            raise SerialException("Can only operate on a valid port handle")

        #~ self._port_handle.ReceivedBytesThreshold = 1

        if self._timeout is None:
            self._port_handle.ReadTimeout = System.IO.Ports.SerialPort.InfiniteTimeout
        else:
            self._port_handle.ReadTimeout = int(self._timeout * 1000)

        # if self._timeout != 0 and self._interCharTimeout is not None:
            # timeouts = (int(self._interCharTimeout * 1000),) + timeouts[1:]

        if self._write_timeout is None:
            self._port_handle.WriteTimeout = System.IO.Ports.SerialPort.InfiniteTimeout
        else:
            self._port_handle.WriteTimeout = int(self._write_timeout * 1000)

        # Setup the connection info.
        try:
            self._port_handle.BaudRate = self._baudrate
        except IOError as e:
            # catch errors from illegal baudrate settings
            raise ValueError(str(e))

        if self._bytesize == FIVEBITS:
            self._port_handle.DataBits = 5
        elif self._bytesize == SIXBITS:
            self._port_handle.DataBits = 6
        elif self._bytesize == SEVENBITS:
            self._port_handle.DataBits = 7
        elif self._bytesize == EIGHTBITS:
            self._port_handle.DataBits = 8
        else:
            raise ValueError("Unsupported number of data bits: %r" % self._bytesize)

        if self._parity == PARITY_NONE:
            self._port_handle.Parity = getattr(System.IO.Ports.Parity, 'None')  # reserved keyword in Py3k
        elif self._parity == PARITY_EVEN:
            self._port_handle.Parity = System.IO.Ports.Parity.Even
        elif self._parity == PARITY_ODD:
            self._port_handle.Parity = System.IO.Ports.Parity.Odd
        elif self._parity == PARITY_MARK:
            self._port_handle.Parity = System.IO.Ports.Parity.Mark
        elif self._parity == PARITY_SPACE:
            self._port_handle.Parity = System.IO.Ports.Parity.Space
        else:
            raise ValueError("Unsupported parity mode: %r" % self._parity)

        if self._stopbits == STOPBITS_ONE:
            self._port_handle.StopBits = System.IO.Ports.StopBits.One
        elif self._stopbits == STOPBITS_ONE_POINT_FIVE:
            self._port_handle.StopBits = System.IO.Ports.StopBits.OnePointFive
        elif self._stopbits == STOPBITS_TWO:
            self._port_handle.StopBits = System.IO.Ports.StopBits.Two
        else:
            raise ValueError("Unsupported number of stop bits: %r" % self._stopbits)

        if self._rtscts and self._xonxoff:
            self._port_handle.Handshake = System.IO.Ports.Handshake.RequestToSendXOnXOff
        elif self._rtscts:
            self._port_handle.Handshake = System.IO.Ports.Handshake.RequestToSend
        elif self._xonxoff:
            self._port_handle.Handshake = System.IO.Ports.Handshake.XOnXOff
        else:
            self._port_handle.Handshake = getattr(System.IO.Ports.Handshake, 'None')   # reserved keyword in Py3k

    #~ def __del__(self):
        #~ self.close()

    def close(self):
        """Close port"""
        if self.is_open:
            if self._port_handle:
                try:
                    self._port_handle.Close()
                except System.IO.Ports.InvalidOperationException:
                    # ignore errors. can happen for unplugged USB serial devices
                    pass
                self._port_handle = None
            self.is_open = False

    #  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -  -

    @property
    def in_waiting(self):
        """Return the number of characters currently in the input buffer."""
        if not self.is_open:
            raise PortNotOpenError()
        return self._port_handle.BytesToRead

    def read(self, size=1):
        """\
        Read size bytes from the serial port. If a timeout is set it may
        return less characters as requested. With no timeout it will block
        until the requested number of bytes is read.
        """
        if not self.is_open:
            raise PortNotOpenError()
        # must use single byte reads as this is the only way to read
        # without applying encodings
        data = bytearray()
        while size:
            try:
                data.append(self._port_handle.ReadByte())
            except System.TimeoutException:
                break
            else:
                size -= 1
        return bytes(data)

    def write(self, data):
        """Output the given string over the serial port."""
        if not self.is_open:
            raise PortNotOpenError()
        #~ if not isinstance(data, (bytes, bytearray)):
            #~ raise TypeError('expected %s or bytearray, got %s' % (bytes, type(data)))
        try:
            # must call overloaded method with byte array argument
            # as this is the only one not applying encodings
            self._port_handle.Write(as_byte_array(data), 0, len(data))
        except System.TimeoutException:
            raise SerialTimeoutException('Write timeout')
        return len(data)

    def reset_input_buffer(self):
        """Clear input buffer, discarding all that is in the buffer."""
        if not self.is_open:
            raise PortNotOpenError()
        self._port_handle.DiscardInBuffer()

    def reset_output_buffer(self):
        """\
        Clear output buffer, aborting the current output and
        discarding all that is in the buffer.
        """
        if not self.is_open:
            raise PortNotOpenError()
        self._port_handle.DiscardOutBuffer()

    def _update_break_state(self):
        """
        Set break: Controls TXD. When active, to transmitting is possible.
        """
        if not self.is_open:
            raise PortNotOpenError()
        self._port_handle.BreakState = bool(self._break_state)

    def _update_rts_state(self):
        """Set terminal status line: Request To Send"""
        if not self.is_open:
            raise PortNotOpenError()
        self._port_handle.RtsEnable = bool(self._rts_state)

    def _update_dtr_state(self):
        """Set terminal status line: Data Terminal Ready"""
        if not self.is_open:
            raise PortNotOpenError()
        self._port_handle.DtrEnable = bool(self._dtr_state)

    @property
    def cts(self):
        """Read terminal status line: Clear To Send"""
        if not self.is_open:
            raise PortNotOpenError()
        return self._port_handle.CtsHolding

    @property
    def dsr(self):
        """Read terminal status line: Data Set Ready"""
        if not self.is_open:
            raise PortNotOpenError()
        return self._port_handle.DsrHolding

    @property
    def ri(self):
        """Read terminal status line: Ring Indicator"""
        if not self.is_open:
            raise PortNotOpenError()
        #~ return self._port_handle.XXX
        return False  # XXX an error would be better

    @property
    def cd(self):
        """Read terminal status line: Carrier Detect"""
        if not self.is_open:
            raise PortNotOpenError()
        return self._port_handle.CDHolding

    # - - platform specific - - - -
    # none



from typing import (
    List,
    Optional,
)
from ._CustomProperty import CustomProperty
from ._Tender import Tender
from ._SaleLine import SaleLine


class Order(object):
    def __init__(
        self,                         # type: Order
        order_id=None,                # type: int
        state=None,                   # type: Optional[str]
        total_amount=None,            # type: Optional[float]
        total_gross=None,             # type: Optional[float]
        discount_amount=None,         # type: Optional[float]
        total_tender=None,            # type: Optional[float]
        total_after_discount=None,    # type: Optional[float]
        sale_lines=None,              # type: Optional[List[SaleLine]]
        custom_properties=None,       # type: Optional[List[CustomProperty]]
        tenders=None                  # type: Optional[List[Tender]]
    ):
        # type: (...) -> None

        self.order_id = order_id
        self.state = state
        self.total_amount = total_amount
        self.total_gross = total_gross
        self.discount_amount = discount_amount
        self.total_tender = total_tender
        self.total_after_discount = total_after_discount
        self.total_gross = total_gross
        self.sale_lines = sale_lines or []
        self.custom_properties = custom_properties or []
        self.tenders = tenders or []

    def get_custom_property(
        self,   # type: Order
        key,    # type: str
    ):
        # type: (...) -> Optional[CustomProperty]

        for custom_property in self.custom_properties:
            if custom_property.key == key:
                return custom_property

        return None



# ### this file stubs are generated by tools/write_pyi.py - do not edit ###
# ### imports are manually managed
from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Collection
from typing import ContextManager
from typing import Dict
from typing import Iterable
from typing import List
from typing import Literal
from typing import Mapping
from typing import MutableMapping
from typing import Optional
from typing import overload
from typing import Sequence
from typing import TextIO
from typing import Tuple
from typing import TYPE_CHECKING
from typing import Union

if TYPE_CHECKING:
    from sqlalchemy.engine.base import Connection
    from sqlalchemy.engine.url import URL
    from sqlalchemy.sql import Executable
    from sqlalchemy.sql.schema import Column
    from sqlalchemy.sql.schema import FetchedValue
    from sqlalchemy.sql.schema import MetaData
    from sqlalchemy.sql.schema import SchemaItem
    from sqlalchemy.sql.type_api import TypeEngine

    from .autogenerate.api import AutogenContext
    from .config import Config
    from .operations.ops import MigrationScript
    from .runtime.migration import _ProxyTransaction
    from .runtime.migration import MigrationContext
    from .runtime.migration import MigrationInfo
    from .script import ScriptDirectory

### end imports ###

def begin_transaction() -> Union[_ProxyTransaction, ContextManager[None]]:
    """Return a context manager that will
    enclose an operation within a "transaction",
    as defined by the environment's offline
    and transactional DDL settings.

    e.g.::

        with context.begin_transaction():
            context.run_migrations()

    :meth:`.begin_transaction` is intended to
    "do the right thing" regardless of
    calling context:

    * If :meth:`.is_transactional_ddl` is ``False``,
      returns a "do nothing" context manager
      which otherwise produces no transactional
      state or directives.
    * If :meth:`.is_offline_mode` is ``True``,
      returns a context manager that will
      invoke the :meth:`.DefaultImpl.emit_begin`
      and :meth:`.DefaultImpl.emit_commit`
      methods, which will produce the string
      directives ``BEGIN`` and ``COMMIT`` on
      the output stream, as rendered by the
      target backend (e.g. SQL Server would
      emit ``BEGIN TRANSACTION``).
    * Otherwise, calls :meth:`sqlalchemy.engine.Connection.begin`
      on the current online connection, which
      returns a :class:`sqlalchemy.engine.Transaction`
      object.  This object demarcates a real
      transaction and is itself a context manager,
      which will roll back if an exception
      is raised.

    Note that a custom ``env.py`` script which
    has more specific transactional needs can of course
    manipulate the :class:`~sqlalchemy.engine.Connection`
    directly to produce transactional state in "online"
    mode.

    """

config: Config

def configure(
    connection: Optional[Connection] = None,
    url: Union[str, URL, None] = None,
    dialect_name: Optional[str] = None,
    dialect_opts: Optional[Dict[str, Any]] = None,
    transactional_ddl: Optional[bool] = None,
    transaction_per_migration: bool = False,
    output_buffer: Optional[TextIO] = None,
    starting_rev: Optional[str] = None,
    tag: Optional[str] = None,
    template_args: Optional[Dict[str, Any]] = None,
    render_as_batch: bool = False,
    target_metadata: Union[MetaData, Sequence[MetaData], None] = None,
    include_name: Optional[
        Callable[
            [
                Optional[str],
                Literal[
                    "schema",
                    "table",
                    "column",
                    "index",
                    "unique_constraint",
                    "foreign_key_constraint",
                ],
                MutableMapping[
                    Literal[
                        "schema_name",
                        "table_name",
                        "schema_qualified_table_name",
                    ],
                    Optional[str],
                ],
            ],
            bool,
        ]
    ] = None,
    include_object: Optional[
        Callable[
            [
                SchemaItem,
                Optional[str],
                Literal[
                    "schema",
                    "table",
                    "column",
                    "index",
                    "unique_constraint",
                    "foreign_key_constraint",
                ],
                bool,
                Optional[SchemaItem],
            ],
            bool,
        ]
    ] = None,
    include_schemas: bool = False,
    process_revision_directives: Optional[
        Callable[
            [
                MigrationContext,
                Union[str, Iterable[Optional[str]], Iterable[str]],
                List[MigrationScript],
            ],
            None,
        ]
    ] = None,
    compare_type: Union[
        bool,
        Callable[
            [
                MigrationContext,
                Column[Any],
                Column[Any],
                TypeEngine,
                TypeEngine,
            ],
            Optional[bool],
        ],
    ] = True,
    compare_server_default: Union[
        bool,
        Callable[
            [
                MigrationContext,
                Column[Any],
                Column[Any],
                Optional[str],
                Optional[FetchedValue],
                Optional[str],
            ],
            Optional[bool],
        ],
    ] = False,
    render_item: Optional[
        Callable[[str, Any, AutogenContext], Union[str, Literal[False]]]
    ] = None,
    literal_binds: bool = False,
    upgrade_token: str = "upgrades",
    downgrade_token: str = "downgrades",
    alembic_module_prefix: str = "op.",
    sqlalchemy_module_prefix: str = "sa.",
    user_module_prefix: Optional[str] = None,
    on_version_apply: Optional[
        Callable[
            [
                MigrationContext,
                MigrationInfo,
                Collection[Any],
                Mapping[str, Any],
            ],
            None,
        ]
    ] = None,
    **kw: Any,
) -> None:
    """Configure a :class:`.MigrationContext` within this
    :class:`.EnvironmentContext` which will provide database
    connectivity and other configuration to a series of
    migration scripts.

    Many methods on :class:`.EnvironmentContext` require that
    this method has been called in order to function, as they
    ultimately need to have database access or at least access
    to the dialect in use.  Those which do are documented as such.

    The important thing needed by :meth:`.configure` is a
    means to determine what kind of database dialect is in use.
    An actual connection to that database is needed only if
    the :class:`.MigrationContext` is to be used in
    "online" mode.

    If the :meth:`.is_offline_mode` function returns ``True``,
    then no connection is needed here.  Otherwise, the
    ``connection`` parameter should be present as an
    instance of :class:`sqlalchemy.engine.Connection`.

    This function is typically called from the ``env.py``
    script within a migration environment.  It can be called
    multiple times for an invocation.  The most recent
    :class:`~sqlalchemy.engine.Connection`
    for which it was called is the one that will be operated upon
    by the next call to :meth:`.run_migrations`.

    General parameters:

    :param connection: a :class:`~sqlalchemy.engine.Connection`
     to use
     for SQL execution in "online" mode.  When present, is also
     used to determine the type of dialect in use.
    :param url: a string database url, or a
     :class:`sqlalchemy.engine.url.URL` object.
     The type of dialect to be used will be derived from this if
     ``connection`` is not passed.
    :param dialect_name: string name of a dialect, such as
     "postgresql", "mssql", etc.
     The type of dialect to be used will be derived from this if
     ``connection`` and ``url`` are not passed.
    :param dialect_opts: dictionary of options to be passed to dialect
     constructor.
    :param transactional_ddl: Force the usage of "transactional"
     DDL on or off;
     this otherwise defaults to whether or not the dialect in
     use supports it.
    :param transaction_per_migration: if True, nest each migration script
     in a transaction rather than the full series of migrations to
     run.
    :param output_buffer: a file-like object that will be used
     for textual output
     when the ``--sql`` option is used to generate SQL scripts.
     Defaults to
     ``sys.stdout`` if not passed here and also not present on
     the :class:`.Config`
     object.  The value here overrides that of the :class:`.Config`
     object.
    :param output_encoding: when using ``--sql`` to generate SQL
     scripts, apply this encoding to the string output.
    :param literal_binds: when using ``--sql`` to generate SQL
     scripts, pass through the ``literal_binds`` flag to the compiler
     so that any literal values that would ordinarily be bound
     parameters are converted to plain strings.

     .. warning:: Dialects can typically only handle simple datatypes
        like strings and numbers for auto-literal generation.  Datatypes
        like dates, intervals, and others may still require manual
        formatting, typically using :meth:`.Operations.inline_literal`.

     .. note:: the ``literal_binds`` flag is ignored on SQLAlchemy
        versions prior to 0.8 where this feature is not supported.

     .. seealso::

        :meth:`.Operations.inline_literal`

    :param starting_rev: Override the "starting revision" argument
     when using ``--sql`` mode.
    :param tag: a string tag for usage by custom ``env.py`` scripts.
     Set via the ``--tag`` option, can be overridden here.
    :param template_args: dictionary of template arguments which
     will be added to the template argument environment when
     running the "revision" command.   Note that the script environment
     is only run within the "revision" command if the --autogenerate
     option is used, or if the option "revision_environment=true"
     is present in the alembic.ini file.

    :param version_table: The name of the Alembic version table.
     The default is ``'alembic_version'``.
    :param version_table_schema: Optional schema to place version
     table within.
    :param version_table_pk: boolean, whether the Alembic version table
     should use a primary key constraint for the "value" column; this
     only takes effect when the table is first created.
     Defaults to True; setting to False should not be necessary and is
     here for backwards compatibility reasons.
    :param on_version_apply: a callable or collection of callables to be
        run for each migration step.
        The callables will be run in the order they are given, once for
        each migration step, after the respective operation has been
        applied but before its transaction is finalized.
        Each callable accepts no positional arguments and the following
        keyword arguments:

        * ``ctx``: the :class:`.MigrationContext` running the migration,
        * ``step``: a :class:`.MigrationInfo` representing the
          step currently being applied,
        * ``heads``: a collection of version strings representing the
          current heads,
        * ``run_args``: the ``**kwargs`` passed to :meth:`.run_migrations`.

    Parameters specific to the autogenerate feature, when
    ``alembic revision`` is run with the ``--autogenerate`` feature:

    :param target_metadata: a :class:`sqlalchemy.schema.MetaData`
     object, or a sequence of :class:`~sqlalchemy.schema.MetaData`
     objects, that will be consulted during autogeneration.
     The tables present in each :class:`~sqlalchemy.schema.MetaData`
     will be compared against
     what is locally available on the target
     :class:`~sqlalchemy.engine.Connection`
     to produce candidate upgrade/downgrade operations.
    :param compare_type: Indicates type comparison behavior during
     an autogenerate
     operation.  Defaults to ``True`` turning on type comparison, which
     has good accuracy on most backends.   See :ref:`compare_types`
     for an example as well as information on other type
     comparison options. Set to ``False`` which disables type
     comparison. A callable can also be passed to provide custom type
     comparison, see :ref:`compare_types` for additional details.

     .. versionchanged:: 1.12.0 The default value of
        :paramref:`.EnvironmentContext.configure.compare_type` has been
        changed to ``True``.

     .. seealso::

        :ref:`compare_types`

        :paramref:`.EnvironmentContext.configure.compare_server_default`

    :param compare_server_default: Indicates server default comparison
     behavior during
     an autogenerate operation.  Defaults to ``False`` which disables
     server default
     comparison.  Set to  ``True`` to turn on server default comparison,
     which has
     varied accuracy depending on backend.

     To customize server default comparison behavior, a callable may
     be specified
     which can filter server default comparisons during an
     autogenerate operation.
     defaults during an autogenerate operation.   The format of this
     callable is::

        def my_compare_server_default(context, inspected_column,
                    metadata_column, inspected_default, metadata_default,
                    rendered_metadata_default):
            # return True if the defaults are different,
            # False if not, or None to allow the default implementation
            # to compare these defaults
            return None

        context.configure(
            # ...
            compare_server_default = my_compare_server_default
        )

     ``inspected_column`` is a dictionary structure as returned by
     :meth:`sqlalchemy.engine.reflection.Inspector.get_columns`, whereas
     ``metadata_column`` is a :class:`sqlalchemy.schema.Column` from
     the local model environment.

     A return value of ``None`` indicates to allow default server default
     comparison
     to proceed.  Note that some backends such as Postgresql actually
     execute
     the two defaults on the database side to compare for equivalence.

     .. seealso::

        :paramref:`.EnvironmentContext.configure.compare_type`

    :param include_name: A callable function which is given
     the chance to return ``True`` or ``False`` for any database reflected
     object based on its name, including database schema names when
     the :paramref:`.EnvironmentContext.configure.include_schemas` flag
     is set to ``True``.

     The function accepts the following positional arguments:

     * ``name``: the name of the object, such as schema name or table name.
       Will be ``None`` when indicating the default schema name of the
       database connection.
     * ``type``: a string describing the type of object; currently
       ``"schema"``, ``"table"``, ``"column"``, ``"index"``,
       ``"unique_constraint"``, or ``"foreign_key_constraint"``
     * ``parent_names``: a dictionary of "parent" object names, that are
       relative to the name being given.  Keys in this dictionary may
       include:  ``"schema_name"``, ``"table_name"`` or
       ``"schema_qualified_table_name"``.

     E.g.::

        def include_name(name, type_, parent_names):
            if type_ == "schema":
                return name in ["schema_one", "schema_two"]
            else:
                return True

        context.configure(
            # ...
            include_schemas = True,
            include_name = include_name
        )

     .. seealso::

        :ref:`autogenerate_include_hooks`

        :paramref:`.EnvironmentContext.configure.include_object`

        :paramref:`.EnvironmentContext.configure.include_schemas`


    :param include_object: A callable function which is given
     the chance to return ``True`` or ``False`` for any object,
     indicating if the given object should be considered in the
     autogenerate sweep.

     The function accepts the following positional arguments:

     * ``object``: a :class:`~sqlalchemy.schema.SchemaItem` object such
       as a :class:`~sqlalchemy.schema.Table`,
       :class:`~sqlalchemy.schema.Column`,
       :class:`~sqlalchemy.schema.Index`
       :class:`~sqlalchemy.schema.UniqueConstraint`,
       or :class:`~sqlalchemy.schema.ForeignKeyConstraint` object
     * ``name``: the name of the object. This is typically available
       via ``object.name``.
     * ``type``: a string describing the type of object; currently
       ``"table"``, ``"column"``, ``"index"``, ``"unique_constraint"``,
       or ``"foreign_key_constraint"``
     * ``reflected``: ``True`` if the given object was produced based on
       table reflection, ``False`` if it's from a local :class:`.MetaData`
       object.
     * ``compare_to``: the object being compared against, if available,
       else ``None``.

     E.g.::

        def include_object(object, name, type_, reflected, compare_to):
            if (type_ == "column" and
                not reflected and
                object.info.get("skip_autogenerate", False)):
                return False
            else:
                return True

        context.configure(
            # ...
            include_object = include_object
        )

     For the use case of omitting specific schemas from a target database
     when :paramref:`.EnvironmentContext.configure.include_schemas` is
     set to ``True``, the :attr:`~sqlalchemy.schema.Table.schema`
     attribute can be checked for each :class:`~sqlalchemy.schema.Table`
     object passed to the hook, however it is much more efficient
     to filter on schemas before reflection of objects takes place
     using the :paramref:`.EnvironmentContext.configure.include_name`
     hook.

     .. seealso::

        :ref:`autogenerate_include_hooks`

        :paramref:`.EnvironmentContext.configure.include_name`

        :paramref:`.EnvironmentContext.configure.include_schemas`

    :param render_as_batch: if True, commands which alter elements
     within a table will be placed under a ``with batch_alter_table():``
     directive, so that batch migrations will take place.

     .. seealso::

        :ref:`batch_migrations`

    :param include_schemas: If True, autogenerate will scan across
     all schemas located by the SQLAlchemy
     :meth:`~sqlalchemy.engine.reflection.Inspector.get_schema_names`
     method, and include all differences in tables found across all
     those schemas.  When using this option, you may want to also
     use the :paramref:`.EnvironmentContext.configure.include_name`
     parameter to specify a callable which
     can filter the tables/schemas that get included.

     .. seealso::

        :ref:`autogenerate_include_hooks`

        :paramref:`.EnvironmentContext.configure.include_name`

        :paramref:`.EnvironmentContext.configure.include_object`

    :param render_item: Callable that can be used to override how
     any schema item, i.e. column, constraint, type,
     etc., is rendered for autogenerate.  The callable receives a
     string describing the type of object, the object, and
     the autogen context.  If it returns False, the
     default rendering method will be used.  If it returns None,
     the item will not be rendered in the context of a Table
     construct, that is, can be used to skip columns or constraints
     within op.create_table()::

        def my_render_column(type_, col, autogen_context):
            if type_ == "column" and isinstance(col, MySpecialCol):
                return repr(col)
            else:
                return False

        context.configure(
            # ...
            render_item = my_render_column
        )

     Available values for the type string include: ``"column"``,
     ``"primary_key"``, ``"foreign_key"``, ``"unique"``, ``"check"``,
     ``"type"``, ``"server_default"``.

     .. seealso::

        :ref:`autogen_render_types`

    :param upgrade_token: When autogenerate completes, the text of the
     candidate upgrade operations will be present in this template
     variable when ``script.py.mako`` is rendered.  Defaults to
     ``upgrades``.
    :param downgrade_token: When autogenerate completes, the text of the
     candidate downgrade operations will be present in this
     template variable when ``script.py.mako`` is rendered.  Defaults to
     ``downgrades``.

    :param alembic_module_prefix: When autogenerate refers to Alembic
     :mod:`alembic.operations` constructs, this prefix will be used
     (i.e. ``op.create_table``)  Defaults to "``op.``".
     Can be ``None`` to indicate no prefix.

    :param sqlalchemy_module_prefix: When autogenerate refers to
     SQLAlchemy
     :class:`~sqlalchemy.schema.Column` or type classes, this prefix
     will be used
     (i.e. ``sa.Column("somename", sa.Integer)``)  Defaults to "``sa.``".
     Can be ``None`` to indicate no prefix.
     Note that when dialect-specific types are rendered, autogenerate
     will render them using the dialect module name, i.e. ``mssql.BIT()``,
     ``postgresql.UUID()``.

    :param user_module_prefix: When autogenerate refers to a SQLAlchemy
     type (e.g. :class:`.TypeEngine`) where the module name is not
     under the ``sqlalchemy`` namespace, this prefix will be used
     within autogenerate.  If left at its default of
     ``None``, the ``__module__`` attribute of the type is used to
     render the import module.   It's a good practice to set this
     and to have all custom types be available from a fixed module space,
     in order to future-proof migration files against reorganizations
     in modules.

     .. seealso::

        :ref:`autogen_module_prefix`

    :param process_revision_directives: a callable function that will
     be passed a structure representing the end result of an autogenerate
     or plain "revision" operation, which can be manipulated to affect
     how the ``alembic revision`` command ultimately outputs new
     revision scripts.   The structure of the callable is::

        def process_revision_directives(context, revision, directives):
            pass

     The ``directives`` parameter is a Python list containing
     a single :class:`.MigrationScript` directive, which represents
     the revision file to be generated.    This list as well as its
     contents may be freely modified to produce any set of commands.
     The section :ref:`customizing_revision` shows an example of
     doing this.  The ``context`` parameter is the
     :class:`.MigrationContext` in use,
     and ``revision`` is a tuple of revision identifiers representing the
     current revision of the database.

     The callable is invoked at all times when the ``--autogenerate``
     option is passed to ``alembic revision``.  If ``--autogenerate``
     is not passed, the callable is invoked only if the
     ``revision_environment`` variable is set to True in the Alembic
     configuration, in which case the given ``directives`` collection
     will contain empty :class:`.UpgradeOps` and :class:`.DowngradeOps`
     collections for ``.upgrade_ops`` and ``.downgrade_ops``.  The
     ``--autogenerate`` option itself can be inferred by inspecting
     ``context.config.cmd_opts.autogenerate``.

     The callable function may optionally be an instance of
     a :class:`.Rewriter` object.  This is a helper object that
     assists in the production of autogenerate-stream rewriter functions.

     .. seealso::

         :ref:`customizing_revision`

         :ref:`autogen_rewriter`

         :paramref:`.command.revision.process_revision_directives`

    Parameters specific to individual backends:

    :param mssql_batch_separator: The "batch separator" which will
     be placed between each statement when generating offline SQL Server
     migrations.  Defaults to ``GO``.  Note this is in addition to the
     customary semicolon ``;`` at the end of each statement; SQL Server
     considers the "batch separator" to denote the end of an
     individual statement execution, and cannot group certain
     dependent operations in one step.
    :param oracle_batch_separator: The "batch separator" which will
     be placed between each statement when generating offline
     Oracle migrations.  Defaults to ``/``.  Oracle doesn't add a
     semicolon between statements like most other backends.

    """

def execute(
    sql: Union[Executable, str], execution_options: Optional[dict] = None
) -> None:
    """Execute the given SQL using the current change context.

    The behavior of :meth:`.execute` is the same
    as that of :meth:`.Operations.execute`.  Please see that
    function's documentation for full detail including
    caveats and limitations.

    This function requires that a :class:`.MigrationContext` has
    first been made available via :meth:`.configure`.

    """

def get_bind() -> Connection:
    """Return the current 'bind'.

    In "online" mode, this is the
    :class:`sqlalchemy.engine.Connection` currently being used
    to emit SQL to the database.

    This function requires that a :class:`.MigrationContext`
    has first been made available via :meth:`.configure`.

    """

def get_context() -> MigrationContext:
    """Return the current :class:`.MigrationContext` object.

    If :meth:`.EnvironmentContext.configure` has not been
    called yet, raises an exception.

    """

def get_head_revision() -> Union[str, Tuple[str, ...], None]:
    """Return the hex identifier of the 'head' script revision.

    If the script directory has multiple heads, this
    method raises a :class:`.CommandError`;
    :meth:`.EnvironmentContext.get_head_revisions` should be preferred.

    This function does not require that the :class:`.MigrationContext`
    has been configured.

    .. seealso:: :meth:`.EnvironmentContext.get_head_revisions`

    """

def get_head_revisions() -> Union[str, Tuple[str, ...], None]:
    """Return the hex identifier of the 'heads' script revision(s).

    This returns a tuple containing the version number of all
    heads in the script directory.

    This function does not require that the :class:`.MigrationContext`
    has been configured.

    """

def get_revision_argument() -> Union[str, Tuple[str, ...], None]:
    """Get the 'destination' revision argument.

    This is typically the argument passed to the
    ``upgrade`` or ``downgrade`` command.

    If it was specified as ``head``, the actual
    version number is returned; if specified
    as ``base``, ``None`` is returned.

    This function does not require that the :class:`.MigrationContext`
    has been configured.

    """

def get_starting_revision_argument() -> Union[str, Tuple[str, ...], None]:
    """Return the 'starting revision' argument,
    if the revision was passed using ``start:end``.

    This is only meaningful in "offline" mode.
    Returns ``None`` if no value is available
    or was configured.

    This function does not require that the :class:`.MigrationContext`
    has been configured.

    """

def get_tag_argument() -> Optional[str]:
    """Return the value passed for the ``--tag`` argument, if any.

    The ``--tag`` argument is not used directly by Alembic,
    but is available for custom ``env.py`` configurations that
    wish to use it; particularly for offline generation scripts
    that wish to generate tagged filenames.

    This function does not require that the :class:`.MigrationContext`
    has been configured.

    .. seealso::

        :meth:`.EnvironmentContext.get_x_argument` - a newer and more
        open ended system of extending ``env.py`` scripts via the command
        line.

    """

@overload
def get_x_argument(as_dictionary: Literal[False]) -> List[str]: ...
@overload
def get_x_argument(as_dictionary: Literal[True]) -> Dict[str, str]: ...
@overload
def get_x_argument(
    as_dictionary: bool = ...,
) -> Union[List[str], Dict[str, str]]:
    """Return the value(s) passed for the ``-x`` argument, if any.

    The ``-x`` argument is an open ended flag that allows any user-defined
    value or values to be passed on the command line, then available
    here for consumption by a custom ``env.py`` script.

    The return value is a list, returned directly from the ``argparse``
    structure.  If ``as_dictionary=True`` is passed, the ``x`` arguments
    are parsed using ``key=value`` format into a dictionary that is
    then returned.

    For example, to support passing a database URL on the command line,
    the standard ``env.py`` script can be modified like this::

        cmd_line_url = context.get_x_argument(
            as_dictionary=True).get('dbname')
        if cmd_line_url:
            engine = create_engine(cmd_line_url)
        else:
            engine = engine_from_config(
                    config.get_section(config.config_ini_section),
                    prefix='sqlalchemy.',
                    poolclass=pool.NullPool)

    This then takes effect by running the ``alembic`` script as::

        alembic -x dbname=postgresql://user:pass@host/dbname upgrade head

    This function does not require that the :class:`.MigrationContext`
    has been configured.

    .. seealso::

        :meth:`.EnvironmentContext.get_tag_argument`

        :attr:`.Config.cmd_opts`

    """

def is_offline_mode() -> bool:
    """Return True if the current migrations environment
    is running in "offline mode".

    This is ``True`` or ``False`` depending
    on the ``--sql`` flag passed.

    This function does not require that the :class:`.MigrationContext`
    has been configured.

    """

def is_transactional_ddl():
    """Return True if the context is configured to expect a
    transactional DDL capable backend.

    This defaults to the type of database in use, and
    can be overridden by the ``transactional_ddl`` argument
    to :meth:`.configure`

    This function requires that a :class:`.MigrationContext`
    has first been made available via :meth:`.configure`.

    """

def run_migrations(**kw: Any) -> None:
    """Run migrations as determined by the current command line
    configuration
    as well as versioning information present (or not) in the current
    database connection (if one is present).

    The function accepts optional ``**kw`` arguments.   If these are
    passed, they are sent directly to the ``upgrade()`` and
    ``downgrade()``
    functions within each target revision file.   By modifying the
    ``script.py.mako`` file so that the ``upgrade()`` and ``downgrade()``
    functions accept arguments, parameters can be passed here so that
    contextual information, usually information to identify a particular
    database in use, can be passed from a custom ``env.py`` script
    to the migration functions.

    This function requires that a :class:`.MigrationContext` has
    first been made available via :meth:`.configure`.

    """

script: ScriptDirectory

def static_output(text: str) -> None:
    """Emit text directly to the "offline" SQL stream.

    Typically this is for emitting comments that
    start with --.  The statement is not treated
    as a SQL execution, no ; or batch separator
    is added, etc.

    """



from pos_model import Order

from _TaxItemFormatter import TaxItemFormatter


class OrderFormatter(object):
    pattern = '<?xml version="1.0" encoding="UTF-8"?>\n'
    pattern += '<Order posId=\"{pos_id}\" orderId=\"{order_id}\">{tax_items}</Order>'

    def __init__(self, tax_item_formatter):
        # type: (TaxItemFormatter) -> None
        self.tax_item_formatter = tax_item_formatter

    def format_order(self, order):
        # type: (Order) -> str
        tax_items_str = ''
        for sale_item in order.sale_items:
            for tax_item in sale_item.taxes:
                tax_items_str += self.tax_item_formatter.format_tax_item(tax_item, sale_item)

        formatted_order = self.pattern.format(
            pos_id=order.pos_id,
            order_id=order.order_id,
            tax_items=tax_items_str)

        return formatted_order



import os
import shutil
import stat
import warnings
from pathlib import Path
from unittest.mock import Mock

import jaraco.path
import pytest

from setuptools import SetuptoolsDeprecationWarning
from setuptools.dist import Distribution

from .textwrap import DALS


def test_directories_in_package_data_glob(tmpdir_cwd):
    """
    Directories matching the glob in package_data should
    not be included in the package data.

    Regression test for #261.
    """
    dist = Distribution(
        dict(
            script_name='setup.py',
            script_args=['build_py'],
            packages=[''],
            package_data={'': ['path/*']},
        )
    )
    os.makedirs('path/subpath')
    dist.parse_command_line()
    dist.run_commands()


def test_recursive_in_package_data_glob(tmpdir_cwd):
    """
    Files matching recursive globs (**) in package_data should
    be included in the package data.

    #1806
    """
    dist = Distribution(
        dict(
            script_name='setup.py',
            script_args=['build_py'],
            packages=[''],
            package_data={'': ['path/**/data']},
        )
    )
    os.makedirs('path/subpath/subsubpath')
    open('path/subpath/subsubpath/data', 'wb').close()

    dist.parse_command_line()
    dist.run_commands()

    assert stat.S_ISREG(
        os.stat('build/lib/path/subpath/subsubpath/data').st_mode
    ), "File is not included"


def test_read_only(tmpdir_cwd):
    """
    Ensure read-only flag is not preserved in copy
    for package modules and package data, as that
    causes problems with deleting read-only files on
    Windows.

    #1451
    """
    dist = Distribution(
        dict(
            script_name='setup.py',
            script_args=['build_py'],
            packages=['pkg'],
            package_data={'pkg': ['data.dat']},
        )
    )
    os.makedirs('pkg')
    open('pkg/__init__.py', 'wb').close()
    open('pkg/data.dat', 'wb').close()
    os.chmod('pkg/__init__.py', stat.S_IREAD)
    os.chmod('pkg/data.dat', stat.S_IREAD)
    dist.parse_command_line()
    dist.run_commands()
    shutil.rmtree('build')


@pytest.mark.xfail(
    'platform.system() == "Windows"',
    reason="On Windows, files do not have executable bits",
    raises=AssertionError,
    strict=True,
)
def test_executable_data(tmpdir_cwd):
    """
    Ensure executable bit is preserved in copy for
    package data, as users rely on it for scripts.

    #2041
    """
    dist = Distribution(
        dict(
            script_name='setup.py',
            script_args=['build_py'],
            packages=['pkg'],
            package_data={'pkg': ['run-me']},
        )
    )
    os.makedirs('pkg')
    open('pkg/__init__.py', 'wb').close()
    open('pkg/run-me', 'wb').close()
    os.chmod('pkg/run-me', 0o700)

    dist.parse_command_line()
    dist.run_commands()

    assert (
        os.stat('build/lib/pkg/run-me').st_mode & stat.S_IEXEC
    ), "Script is not executable"


EXAMPLE_WITH_MANIFEST = {
    "setup.cfg": DALS(
        """
        [metadata]
        name = mypkg
        version = 42

        [options]
        include_package_data = True
        packages = find:

        [options.packages.find]
        exclude = *.tests*
        """
    ),
    "mypkg": {
        "__init__.py": "",
        "resource_file.txt": "",
        "tests": {
            "__init__.py": "",
            "test_mypkg.py": "",
            "test_file.txt": "",
        },
    },
    "MANIFEST.in": DALS(
        """
        global-include *.py *.txt
        global-exclude *.py[cod]
        prune dist
        prune build
        prune *.egg-info
        """
    ),
}


def test_excluded_subpackages(tmpdir_cwd):
    jaraco.path.build(EXAMPLE_WITH_MANIFEST)
    dist = Distribution({"script_name": "%PEP 517%"})
    dist.parse_config_files()

    build_py = dist.get_command_obj("build_py")

    msg = r"Python recognizes 'mypkg\.tests' as an importable package"
    with pytest.warns(SetuptoolsDeprecationWarning, match=msg):
        # TODO: To fix #3260 we need some transition period to deprecate the
        # existing behavior of `include_package_data`. After the transition, we
        # should remove the warning and fix the behaviour.

        if os.getenv("SETUPTOOLS_USE_DISTUTILS") == "stdlib":
            # pytest.warns reset the warning filter temporarily
            # https://github.com/pytest-dev/pytest/issues/4011#issuecomment-423494810
            warnings.filterwarnings(
                "ignore",
                "'encoding' argument not specified",
                module="distutils.text_file",
                # This warning is already fixed in pypa/distutils but not in stdlib
            )

        build_py.finalize_options()
        build_py.run()

    build_dir = Path(dist.get_command_obj("build_py").build_lib)
    assert (build_dir / "mypkg/__init__.py").exists()
    assert (build_dir / "mypkg/resource_file.txt").exists()

    # Setuptools is configured to ignore `mypkg.tests`, therefore the following
    # files/dirs should not be included in the distribution.
    for f in [
        "mypkg/tests/__init__.py",
        "mypkg/tests/test_mypkg.py",
        "mypkg/tests/test_file.txt",
        "mypkg/tests",
    ]:
        with pytest.raises(AssertionError):
            # TODO: Enforce the following assertion once #3260 is fixed
            # (remove context manager and the following xfail).
            assert not (build_dir / f).exists()

    pytest.xfail("#3260")


@pytest.mark.filterwarnings("ignore::setuptools.SetuptoolsDeprecationWarning")
def test_existing_egg_info(tmpdir_cwd, monkeypatch):
    """When provided with the ``existing_egg_info_dir`` attribute, build_py should not
    attempt to run egg_info again.
    """
    # == Pre-condition ==
    # Generate an egg-info dir
    jaraco.path.build(EXAMPLE_WITH_MANIFEST)
    dist = Distribution({"script_name": "%PEP 517%"})
    dist.parse_config_files()
    assert dist.include_package_data

    egg_info = dist.get_command_obj("egg_info")
    dist.run_command("egg_info")
    egg_info_dir = next(Path(egg_info.egg_base).glob("*.egg-info"))
    assert egg_info_dir.is_dir()

    # == Setup ==
    build_py = dist.get_command_obj("build_py")
    build_py.finalize_options()
    egg_info = dist.get_command_obj("egg_info")
    egg_info_run = Mock(side_effect=egg_info.run)
    monkeypatch.setattr(egg_info, "run", egg_info_run)

    # == Remove caches ==
    # egg_info is called when build_py looks for data_files, which gets cached.
    # We need to ensure it is not cached yet, otherwise it may impact on the tests
    build_py.__dict__.pop('data_files', None)
    dist.reinitialize_command(egg_info)

    # == Sanity check ==
    # Ensure that if existing_egg_info is not given, build_py attempts to run egg_info
    build_py.existing_egg_info_dir = None
    build_py.run()
    egg_info_run.assert_called()

    # == Remove caches ==
    egg_info_run.reset_mock()
    build_py.__dict__.pop('data_files', None)
    dist.reinitialize_command(egg_info)

    # == Actual test ==
    # Ensure that if existing_egg_info_dir is given, egg_info doesn't run
    build_py.existing_egg_info_dir = egg_info_dir
    build_py.run()
    egg_info_run.assert_not_called()
    assert build_py.data_files

    # Make sure the list of outputs is actually OK
    outputs = map(lambda x: x.replace(os.sep, "/"), build_py.get_outputs())
    assert outputs
    example = str(Path(build_py.build_lib, "mypkg/__init__.py")).replace(os.sep, "/")
    assert example in outputs


EXAMPLE_ARBITRARY_MAPPING = {
    "pyproject.toml": DALS(
        """
        [project]
        name = "mypkg"
        version = "42"

        [tool.setuptools]
        packages = ["mypkg", "mypkg.sub1", "mypkg.sub2", "mypkg.sub2.nested"]

        [tool.setuptools.package-dir]
        "" = "src"
        "mypkg.sub2" = "src/mypkg/_sub2"
        "mypkg.sub2.nested" = "other"
        """
    ),
    "src": {
        "mypkg": {
            "__init__.py": "",
            "resource_file.txt": "",
            "sub1": {
                "__init__.py": "",
                "mod1.py": "",
            },
            "_sub2": {
                "mod2.py": "",
            },
        },
    },
    "other": {
        "__init__.py": "",
        "mod3.py": "",
    },
    "MANIFEST.in": DALS(
        """
        global-include *.py *.txt
        global-exclude *.py[cod]
        """
    ),
}


def test_get_outputs(tmpdir_cwd):
    jaraco.path.build(EXAMPLE_ARBITRARY_MAPPING)
    dist = Distribution({"script_name": "%test%"})
    dist.parse_config_files()

    build_py = dist.get_command_obj("build_py")
    build_py.editable_mode = True
    build_py.ensure_finalized()
    build_lib = build_py.build_lib.replace(os.sep, "/")
    outputs = {x.replace(os.sep, "/") for x in build_py.get_outputs()}
    assert outputs == {
        f"{build_lib}/mypkg/__init__.py",
        f"{build_lib}/mypkg/resource_file.txt",
        f"{build_lib}/mypkg/sub1/__init__.py",
        f"{build_lib}/mypkg/sub1/mod1.py",
        f"{build_lib}/mypkg/sub2/mod2.py",
        f"{build_lib}/mypkg/sub2/nested/__init__.py",
        f"{build_lib}/mypkg/sub2/nested/mod3.py",
    }
    mapping = {
        k.replace(os.sep, "/"): v.replace(os.sep, "/")
        for k, v in build_py.get_output_mapping().items()
    }
    assert mapping == {
        f"{build_lib}/mypkg/__init__.py": "src/mypkg/__init__.py",
        f"{build_lib}/mypkg/resource_file.txt": "src/mypkg/resource_file.txt",
        f"{build_lib}/mypkg/sub1/__init__.py": "src/mypkg/sub1/__init__.py",
        f"{build_lib}/mypkg/sub1/mod1.py": "src/mypkg/sub1/mod1.py",
        f"{build_lib}/mypkg/sub2/mod2.py": "src/mypkg/_sub2/mod2.py",
        f"{build_lib}/mypkg/sub2/nested/__init__.py": "other/__init__.py",
        f"{build_lib}/mypkg/sub2/nested/mod3.py": "other/mod3.py",
    }


class TestTypeInfoFiles:
    PYPROJECTS = {
        "default_pyproject": DALS(
            """
            [project]
            name = "foo"
            version = "1"
            """
        ),
        "dont_include_package_data": DALS(
            """
            [project]
            name = "foo"
            version = "1"

            [tool.setuptools]
            include-package-data = false
            """
        ),
        "exclude_type_info": DALS(
            """
            [project]
            name = "foo"
            version = "1"

            [tool.setuptools]
            include-package-data = false

            [tool.setuptools.exclude-package-data]
            "*" = ["py.typed", "*.pyi"]
            """
        ),
    }

    EXAMPLES = {
        "simple_namespace": {
            "directory_structure": {
                "foo": {
                    "bar.pyi": "",
                    "py.typed": "",
                    "__init__.py": "",
                }
            },
            "expected_type_files": {"foo/bar.pyi", "foo/py.typed"},
        },
        "nested_inside_namespace": {
            "directory_structure": {
                "foo": {
                    "bar": {
                        "py.typed": "",
                        "mod.pyi": "",
                    }
                }
            },
            "expected_type_files": {"foo/bar/mod.pyi", "foo/bar/py.typed"},
        },
        "namespace_nested_inside_regular": {
            "directory_structure": {
                "foo": {
                    "namespace": {
                        "foo.pyi": "",
                    },
                    "__init__.pyi": "",
                    "py.typed": "",
                }
            },
            "expected_type_files": {
                "foo/namespace/foo.pyi",
                "foo/__init__.pyi",
                "foo/py.typed",
            },
        },
    }

    @pytest.mark.parametrize(
        "pyproject",
        [
            "default_pyproject",
            pytest.param(
                "dont_include_package_data",
                marks=pytest.mark.xfail(reason="pypa/setuptools#4350"),
            ),
        ],
    )
    @pytest.mark.parametrize("example", EXAMPLES.keys())
    def test_type_files_included_by_default(self, tmpdir_cwd, pyproject, example):
        structure = {
            **self.EXAMPLES[example]["directory_structure"],
            "pyproject.toml": self.PYPROJECTS[pyproject],
        }
        expected_type_files = self.EXAMPLES[example]["expected_type_files"]
        jaraco.path.build(structure)

        build_py = get_finalized_build_py()
        outputs = get_outputs(build_py)
        assert expected_type_files <= outputs

    @pytest.mark.parametrize("pyproject", ["exclude_type_info"])
    @pytest.mark.parametrize("example", EXAMPLES.keys())
    def test_type_files_can_be_excluded(self, tmpdir_cwd, pyproject, example):
        structure = {
            **self.EXAMPLES[example]["directory_structure"],
            "pyproject.toml": self.PYPROJECTS[pyproject],
        }
        expected_type_files = self.EXAMPLES[example]["expected_type_files"]
        jaraco.path.build(structure)

        build_py = get_finalized_build_py()
        outputs = get_outputs(build_py)
        assert expected_type_files.isdisjoint(outputs)

    def test_stub_only_package(self, tmpdir_cwd):
        structure = {
            "pyproject.toml": DALS(
                """
                [project]
                name = "foo-stubs"
                version = "1"
                """
            ),
            "foo-stubs": {"__init__.pyi": "", "bar.pyi": ""},
        }
        expected_type_files = {"foo-stubs/__init__.pyi", "foo-stubs/bar.pyi"}
        jaraco.path.build(structure)

        build_py = get_finalized_build_py()
        outputs = get_outputs(build_py)
        assert expected_type_files <= outputs


def get_finalized_build_py(script_name="%build_py-test%"):
    dist = Distribution({"script_name": script_name})
    dist.parse_config_files()
    build_py = dist.get_command_obj("build_py")
    build_py.finalize_options()
    return build_py


def get_outputs(build_py):
    build_dir = Path(build_py.build_lib)
    return {
        os.path.relpath(x, build_dir).replace(os.sep, "/")
        for x in build_py.get_outputs()
    }



# Copyright (C) Dnspython Contributors, see LICENSE for text of ISC license

# Copyright (C) 2011 Nominum, Inc.
#
# Permission to use, copy, modify, and distribute this software and its
# documentation for any purpose with or without fee is hereby granted,
# provided that the above copyright notice and this permission notice
# appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND NOMINUM DISCLAIMS ALL WARRANTIES
# WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL NOMINUM BE LIABLE FOR
# ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
# WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
# ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT
# OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

"""Hashing backwards compatibility wrapper"""

import hashlib
import warnings

warnings.warn(
    "dns.hash module will be removed in future versions. Please use hashlib instead.",
    DeprecationWarning)

hashes = {}
hashes['MD5'] = hashlib.md5
hashes['SHA1'] = hashlib.sha1
hashes['SHA224'] = hashlib.sha224
hashes['SHA256'] = hashlib.sha256
hashes['SHA384'] = hashlib.sha384
hashes['SHA512'] = hashlib.sha512


def get(algorithm):
    return hashes[algorithm.upper()]



from abc import abstractmethod
from typing import (
    Any,
    Dict,
    List,
)
from xml.etree import cElementTree as eTree

from common import FiscalOrderTotals
from common.builder_functions import (
    format_datetime,
    optional,
    pad_access_key,
)
from nreports.builders import DataBuilder
from nreports_api.shared import JSON


class BaseFiscalSat(DataBuilder):

    @abstractmethod
    def build_data(
        self,
        data_to_build: Dict[str, JSON],
        **kwargs: Any,
    ) -> Dict[str, JSON]:

        raise NotImplementedError()

    def _build_header(
        self,
        data_to_build: JSON,
        fiscal_xml: eTree.Element,
    ) -> None:

        data_to_build["STORE_NAME"] = fiscal_xml.find(".//emit/xNome").text
        data_to_build["STORE_UF"] = fiscal_xml.find(".//ide/cUF").text
        data_to_build["STORE_STATE_INSCRIPTION"] = fiscal_xml.find(".//emit/IE").text
        data_to_build["STORE_DOC"] = fiscal_xml.find(".//emit/CNPJ").text
        data_to_build["STORE_ADDRESS"] = fiscal_xml.find(".//enderEmit/xLgr").text
        data_to_build["STORE_NUMBER"] = optional(element=fiscal_xml.find(".//enderEmit/nro")) or ""
        data_to_build["STORE_COMPLEMENT"] = optional(element=fiscal_xml.find(".//enderEmit/xCpl")) or ""
        data_to_build["STORE_NEIGHBORHOOD"] = fiscal_xml.find(".//enderEmit/xBairro").text
        data_to_build["STORE_CITY"] = fiscal_xml.find(".//enderEmit/xMun").text
        data_to_build["STORE_ZIP_CODE"] = fiscal_xml.find(".//emit/enderEmit/CEP").text

    @staticmethod
    def _build_body(
        data_to_build: JSON,
        fiscal_xml: eTree.Element,
        fiscal_order_totals: FiscalOrderTotals,
        customer_document: str,
        products: List[Dict[str, str]],
        tenders: List[Dict[str, str]],
    ) -> None:

        data_to_build["ORDER_ITEMS_QTY"] = len(fiscal_xml.findall(".//det"))
        data_to_build["ORDER_TOTAL_GROSS"] = fiscal_order_totals.total_value
        data_to_build["ORDER_DISCOUNT"] = fiscal_order_totals.total_discount
        data_to_build["ORDER_DUE_AMOUNT"] = fiscal_order_totals.total_paid
        data_to_build["ORDER_CHANGE_AMOUNT"] = fiscal_order_totals.amount_change
        data_to_build["CUSTOMER_DOCUMENT"] = customer_document
        data_to_build["ORDER_ADDITION"] = fiscal_order_totals.total_addition or None

        data_to_build["ITEM_PRODUCT"] = products
        data_to_build["TENDER"] = tenders

    def _build_footer(
        self,
        data_to_build: JSON,
        fiscal_xml: eTree.Element,
    ) -> None:

        sat_emission_date = fiscal_xml.find(".//ide/dEmi").text
        sat_emission_time = fiscal_xml.find(".//ide/hEmi").text
        sat_key = fiscal_xml.find(".//infCFe").attrib["Id"][3:]

        data_to_build["SAT_DOC_NUMBER"] = fiscal_xml.find(".//ide/nCFe").text
        data_to_build["SAT_SERIE_NUMBER"] = fiscal_xml.find(".//ide/nserieSAT").text
        data_to_build["SAT_ACCESS_KEY"] = pad_access_key(access_key=sat_key)
        data_to_build["SAT_EMISSION_DATETIME"] = format_datetime(
            datetime_string=sat_emission_date + sat_emission_time,
            input_mask="%Y%m%d%H%M%S",
        )
        inf_cpl = optional(element=fiscal_xml.find(".//infCpl")) or ""
        inf_cpl = "\n\n".join(inf_cpl.split("///"))
        data_to_build["SAT_ADDITIONAL_INFORMATION"] = inf_cpl
        data_to_build["SAT_BARCODE"] = sat_key
        cpf = optional(element=fiscal_xml.find(".//dest/CPF"))
        cnpj = optional(element=fiscal_xml.find(".//dest/CNPJ"))
        data_to_build["SAT_QRCODE"] = "|".join([
            sat_key,
            sat_emission_date + sat_emission_time,
            fiscal_xml.find(".//total/vCFe").text,
            cpf or cnpj or "",
            fiscal_xml.find(".//ide/assinaturaQRCODE").text,
        ])



# Note: This file could use Python 3-only syntax, but at the moment this breaks
# the coverage job on Python 2. Until we make it so that coverage can ignore
# this file only on Python 2, we'll have to stick to 2/3-compatible syntax.
try:
    import zoneinfo
except ImportError:
    from backports import zoneinfo

import datetime

UTC = datetime.timezone.utc


def get_timezone(key):
    try:
        return zoneinfo.ZoneInfo(key)
    except (ValueError, OSError):
        # TODO: Use `from e` when this file can use Python 3 syntax
        raise KeyError(key)


def get_timezone_file(f, key=None):
    return zoneinfo.ZoneInfo.from_file(f, key=key)


def get_fixed_offset_zone(offset):
    return datetime.timezone(datetime.timedelta(minutes=offset))


def is_imaginary(dt):
    dt_rt = dt.astimezone(UTC).astimezone(dt.tzinfo)

    return not (dt == dt_rt)


def is_ambiguous(dt):
    if is_imaginary(dt):
        return False

    wall_0 = dt
    wall_1 = dt.replace(fold=not dt.fold)

    # Ambiguous datetimes can only exist if the offset changes, so we don't
    # actually have to check whether dst() or tzname() are different.
    same_offset = wall_0.utcoffset() == wall_1.utcoffset()

    return not same_offset


def enfold(dt, fold=1):
    if dt.fold != fold:
        return dt.replace(fold=fold)
    else:
        return dt


def get_fold(dt):
    return dt.fold



import http.client
import urllib.error
import urllib.request
from inspect import cleandoc

import pytest

import setuptools.package_index

import distutils.errors


class TestPackageIndex:
    def test_regex(self):
        hash_url = 'http://other_url?:action=show_md5&amp;'
        hash_url += 'digest=0123456789abcdef0123456789abcdef'
        doc = """
            <a href="http://some_url">Name</a>
            (<a title="MD5 hash"
            href="{hash_url}">md5</a>)
        """.lstrip().format(**locals())
        assert setuptools.package_index.PYPI_MD5.match(doc)

    def test_bad_url_bad_port(self):
        index = setuptools.package_index.PackageIndex()
        url = 'http://127.0.0.1:0/nonesuch/test_package_index'
        try:
            v = index.open_url(url)
        except Exception as exc:
            assert url in str(exc)
        else:
            assert isinstance(v, urllib.error.HTTPError)

    def test_bad_url_typo(self):
        # issue 16
        # easy_install inquant.contentmirror.plone breaks because of a typo
        # in its home URL
        index = setuptools.package_index.PackageIndex(hosts=('www.example.com',))

        url = (
            'url:%20https://svn.plone.org/svn'
            '/collective/inquant.contentmirror.plone/trunk'
        )
        try:
            v = index.open_url(url)
        except Exception as exc:
            assert url in str(exc)
        else:
            assert isinstance(v, urllib.error.HTTPError)

    def test_bad_url_bad_status_line(self):
        index = setuptools.package_index.PackageIndex(hosts=('www.example.com',))

        def _urlopen(*args):
            raise http.client.BadStatusLine('line')

        index.opener = _urlopen
        url = 'http://example.com'
        try:
            index.open_url(url)
        except Exception as exc:
            assert 'line' in str(exc)
        else:
            raise AssertionError('Should have raise here!')

    def test_bad_url_double_scheme(self):
        """
        A bad URL with a double scheme should raise a DistutilsError.
        """
        index = setuptools.package_index.PackageIndex(hosts=('www.example.com',))

        # issue 20
        url = 'http://http://svn.pythonpaste.org/Paste/wphp/trunk'
        try:
            index.open_url(url)
        except distutils.errors.DistutilsError as error:
            msg = str(error)
            assert (
                'nonnumeric port' in msg
                or 'getaddrinfo failed' in msg
                or 'Name or service not known' in msg
            )
            return
        raise RuntimeError("Did not raise")

    def test_url_ok(self):
        index = setuptools.package_index.PackageIndex(hosts=('www.example.com',))
        url = 'file:///tmp/test_package_index'
        assert index.url_ok(url, True)

    def test_parse_bdist_wininst(self):
        parse = setuptools.package_index.parse_bdist_wininst

        actual = parse('reportlab-2.5.win32-py2.4.exe')
        expected = 'reportlab-2.5', '2.4', 'win32'
        assert actual == expected

        actual = parse('reportlab-2.5.win32.exe')
        expected = 'reportlab-2.5', None, 'win32'
        assert actual == expected

        actual = parse('reportlab-2.5.win-amd64-py2.7.exe')
        expected = 'reportlab-2.5', '2.7', 'win-amd64'
        assert actual == expected

        actual = parse('reportlab-2.5.win-amd64.exe')
        expected = 'reportlab-2.5', None, 'win-amd64'
        assert actual == expected

    def test__vcs_split_rev_from_url(self):
        """
        Test the basic usage of _vcs_split_rev_from_url
        """
        vsrfu = setuptools.package_index.PackageIndex._vcs_split_rev_from_url
        url, rev = vsrfu('https://example.com/bar@2995')
        assert url == 'https://example.com/bar'
        assert rev == '2995'

    def test_local_index(self, tmpdir):
        """
        local_open should be able to read an index from the file system.
        """
        index_file = tmpdir / 'index.html'
        with index_file.open('w') as f:
            f.write('<div>content</div>')
        url = 'file:' + urllib.request.pathname2url(str(tmpdir)) + '/'
        res = setuptools.package_index.local_open(url)
        assert 'content' in res.read()

    def test_egg_fragment(self):
        """
        EGG fragments must comply to PEP 440
        """
        epoch = [
            '',
            '1!',
        ]
        releases = [
            '0',
            '0.0',
            '0.0.0',
        ]
        pre = [
            'a0',
            'b0',
            'rc0',
        ]
        post = ['.post0']
        dev = [
            '.dev0',
        ]
        local = [
            ('', ''),
            ('+ubuntu.0', '+ubuntu.0'),
            ('+ubuntu-0', '+ubuntu.0'),
            ('+ubuntu_0', '+ubuntu.0'),
        ]
        versions = [
            [''.join([e, r, p, loc]) for loc in locs]
            for e in epoch
            for r in releases
            for p in sum([pre, post, dev], [''])
            for locs in local
        ]
        for v, vc in versions:
            dists = list(
                setuptools.package_index.distros_for_url(
                    'http://example.com/example-foo.zip#egg=example-foo-' + v
                )
            )
            assert dists[0].version == ''
            assert dists[1].version == vc

    def test_download_git_with_rev(self, tmp_path, fp):
        url = 'git+https://github.example/group/project@master#egg=foo'
        index = setuptools.package_index.PackageIndex()

        expected_dir = tmp_path / 'project@master'
        fp.register([
            'git',
            'clone',
            '--quiet',
            'https://github.example/group/project',
            expected_dir,
        ])
        fp.register(['git', '-C', expected_dir, 'checkout', '--quiet', 'master'])

        result = index.download(url, tmp_path)

        assert result == str(expected_dir)
        assert len(fp.calls) == 2

    def test_download_git_no_rev(self, tmp_path, fp):
        url = 'git+https://github.example/group/project#egg=foo'
        index = setuptools.package_index.PackageIndex()

        expected_dir = tmp_path / 'project'
        fp.register([
            'git',
            'clone',
            '--quiet',
            'https://github.example/group/project',
            expected_dir,
        ])
        index.download(url, tmp_path)

    def test_download_svn(self, tmp_path):
        url = 'svn+https://svn.example/project#egg=foo'
        index = setuptools.package_index.PackageIndex()

        msg = r".*SVN download is not supported.*"
        with pytest.raises(distutils.errors.DistutilsError, match=msg):
            index.download(url, tmp_path)


class TestContentCheckers:
    def test_md5(self):
        checker = setuptools.package_index.HashChecker.from_url(
            'http://foo/bar#md5=f12895fdffbd45007040d2e44df98478'
        )
        checker.feed('You should probably not be using MD5'.encode('ascii'))
        assert checker.hash.hexdigest() == 'f12895fdffbd45007040d2e44df98478'
        assert checker.is_valid()

    def test_other_fragment(self):
        "Content checks should succeed silently if no hash is present"
        checker = setuptools.package_index.HashChecker.from_url(
            'http://foo/bar#something%20completely%20different'
        )
        checker.feed('anything'.encode('ascii'))
        assert checker.is_valid()

    def test_blank_md5(self):
        "Content checks should succeed if a hash is empty"
        checker = setuptools.package_index.HashChecker.from_url('http://foo/bar#md5=')
        checker.feed('anything'.encode('ascii'))
        assert checker.is_valid()

    def test_get_hash_name_md5(self):
        checker = setuptools.package_index.HashChecker.from_url(
            'http://foo/bar#md5=f12895fdffbd45007040d2e44df98478'
        )
        assert checker.hash_name == 'md5'

    def test_report(self):
        checker = setuptools.package_index.HashChecker.from_url(
            'http://foo/bar#md5=f12895fdffbd45007040d2e44df98478'
        )
        rep = checker.report(lambda x: x, 'My message about %s')
        assert rep == 'My message about md5'


class TestPyPIConfig:
    def test_percent_in_password(self, tmp_home_dir):
        pypirc = tmp_home_dir / '.pypirc'
        pypirc.write_text(
            cleandoc(
                """
                [pypi]
                repository=https://pypi.org
                username=jaraco
                password=pity%
                """
            ),
            encoding="utf-8",
        )
        cfg = setuptools.package_index.PyPIConfig()
        cred = cfg.creds_by_repository['https://pypi.org']
        assert cred.username == 'jaraco'
        assert cred.password == 'pity%'


@pytest.mark.timeout(1)
def test_REL_DoS():
    """
    REL should not hang on a contrived attack string.
    """
    setuptools.package_index.REL.search('< rel=' + ' ' * 2**12)



# postal_code.py - functions for handling Spanish postal code numbers
# coding: utf-8
#
# Copyright (C) 2023 Víctor Ramos
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""Postcode (the Spanish postal code).

The Spanish postal code consists of five digits where the first two digits,
ranging 01 to 52, correspond either to one of the 50 provinces of Spain or to
one of the two autonomous cities on the African coast.

More information:

* https://en.wikipedia.org/wiki/Postal_codes_in_Spain

>>> validate('01000')
'01000'
>>> validate('52000')
'52000'
>>> validate('00000')
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> validate('53000')
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> validate('99999')
Traceback (most recent call last):
    ...
InvalidComponent: ...
>>> validate('5200')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> validate('520000')
Traceback (most recent call last):
    ...
InvalidLength: ...
"""


from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation."""
    return clean(number, ' ').strip()


def validate(number):
    """Check if the number provided is a valid postal code."""
    number = compact(number)
    if len(number) != 5:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    if not '01' <= number[:2] <= '52':
        raise InvalidComponent()
    return number


def is_valid(number):
    """Check if the number provided is a valid postal code."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False



from __future__ import absolute_import
import sys

assert sys.version_info[0] < 3

from httplib import *
from httplib import HTTPMessage

# These constants aren't included in __all__ in httplib.py:

from httplib import (HTTP_PORT,
                     HTTPS_PORT,

                     _CS_IDLE,
                     _CS_REQ_STARTED,
                     _CS_REQ_SENT,

                     CONTINUE,
                     SWITCHING_PROTOCOLS,
                     PROCESSING,

                     OK,
                     CREATED,
                     ACCEPTED,
                     NON_AUTHORITATIVE_INFORMATION,
                     NO_CONTENT,
                     RESET_CONTENT,
                     PARTIAL_CONTENT,
                     MULTI_STATUS,
                     IM_USED,

                     MULTIPLE_CHOICES,
                     MOVED_PERMANENTLY,
                     FOUND,
                     SEE_OTHER,
                     NOT_MODIFIED,
                     USE_PROXY,
                     TEMPORARY_REDIRECT,

                     BAD_REQUEST,
                     UNAUTHORIZED,
                     PAYMENT_REQUIRED,
                     FORBIDDEN,
                     NOT_FOUND,
                     METHOD_NOT_ALLOWED,
                     NOT_ACCEPTABLE,
                     PROXY_AUTHENTICATION_REQUIRED,
                     REQUEST_TIMEOUT,
                     CONFLICT,
                     GONE,
                     LENGTH_REQUIRED,
                     PRECONDITION_FAILED,
                     REQUEST_ENTITY_TOO_LARGE,
                     REQUEST_URI_TOO_LONG,
                     UNSUPPORTED_MEDIA_TYPE,
                     REQUESTED_RANGE_NOT_SATISFIABLE,
                     EXPECTATION_FAILED,
                     UNPROCESSABLE_ENTITY,
                     LOCKED,
                     FAILED_DEPENDENCY,
                     UPGRADE_REQUIRED,

                     INTERNAL_SERVER_ERROR,
                     NOT_IMPLEMENTED,
                     BAD_GATEWAY,
                     SERVICE_UNAVAILABLE,
                     GATEWAY_TIMEOUT,
                     HTTP_VERSION_NOT_SUPPORTED,
                     INSUFFICIENT_STORAGE,
                     NOT_EXTENDED,

                     MAXAMOUNT,
                    )

# These are not available on Python 2.6.x:
try:
    from httplib import LineTooLong, LineAndFileWrapper
except ImportError:
    pass

# These may not be available on all versions of Python 2.6.x or 2.7.x
try:
    from httplib import (
                         _MAXLINE,
                         _MAXHEADERS,
                         _is_legal_header_name,
                         _is_illegal_header_value,
                         _METHODS_EXPECTING_BODY
                        )
except ImportError:
    pass
